from teneto.communitydetection.optimization.simulated_annealing import simulated_annealing
import numpy as np
def calc_modularityQ(netIn,clusters, param=None):

    """
    :out: is calc_Q being used as a modularity or cost funciton. If cost, return -calc_Q
    """
    if param == None:
        param={}
        param['out'] = 'cost'

    D = np.sum(netIn,axis=1)
    L = np.sum(np.triu(netIn,k=1))
    Q=np.zeros(1)
    for C in np.unique(clusters):
        Cind = np.where(clusters==C)
        if len(Cind[0])==1:
            l=0
        else:
            Cnet=netIn[Cind][:,Cind].squeeze()
            l=np.sum(np.triu(Cnet,k=1))
        d_s = np.sum(D[Cind])
        Q_s = (l/L)-np.power(d_s/(2*L),2)
        Q=np.sum([Q,Q_s])
    if param['out'] == 'cost':
        return -Q
    elif param['out'] == 'modularity':
        return Q


def calc_modularityQ_posneg(netIn ,clusters,param={}):

    """
    :netInput: 2D network
    :param: dictionary containing
        :'out': = 'modularity' or 'cost' (default).
        :'lambda': [0-1] Lower lambda = lower weight to positive edges. 0.5 (default) = equal to both
    """

    if 'out' not in param:
        param['out'] = 'cost'
    if 'lambda' not in param:
        param['lambda'] = 0.5

    netInP = np.array(netIn)
    netInP[netInP<0] = 0
    netInN = np.array(netIn)*-1
    netInN[netInN<0] = 0

    Qp = calc_modularityQ(netInP,clusters,param)
    Qn = calc_modularityQ(netInN,clusters,param)
    Q = (1-param['lambda'])*Qp-param['lambda']*(Qn)
    return Q

def girvan_newman(netIn,optimization='SA',Cnprior=1,repeat=1,cooldown = 0.995, f=0.5,Tstop=10e-100):

    """

    girvan_newman clustering is an optimization technique.

    **PARAMETERS**

    :netIn: 2 dimensional network (n x n).
    :optimization: 'SA' (only one avilable at present.)
    :Cnprior: Prior on number of clusters believes to exists.
    :repeat: how many times to repeat
    :cooldown: reduction in temeprature for each itteration. Decrase for quicker perfromance (and less exploration).
    :f: proportion of nodes shuffled each iteration (also how many merge/splitting attempts per round). Increase for more neighbours.
    :Tstop: terminate optimization. Increase number for quicker performance.

    """
    #No specified parameters needed here.
    Qparams={}
    Qparams['out']='cost'
    if optimization =='SA':
        C,cost=simulated_annealing(netIn,calc_modularityQ,Qparams,Cnprior=Cnprior,repeat=repeat,cooldown = cooldown, f=f,Tstop=Tstop)

    return C




def traag_bruggeman(netIn,optimization='SA',Qparams={},Cnprior=1,repeat=1,cooldown = 0.995, f=0.5,Tstop=10e-100):

    """
    Good for detecting modularity where negative clusters may be present

    READ MORE
    Traag, V. A., & Bruggeman, J. (2009). Community detection in networks with positive and negative links. Physical Review E - Statistical, Nonlinear, and Soft Matter Physics, 80(3), 1–6. http://doi.org/10.1103/PhysRevE.80.036115

    """

    if 'out' not in Qparams:
        Qparams['out']='cost'
    if 'lambda' not in Qparams:
        Qparams['lambda']=0.5

    if optimization =='SA':
        C,cost=simulated_annealing(netIn,calc_modularityQ_posneg,Qparams,Cnprior=Cnprior,repeat=repeat,cooldown = cooldown, f=f,Tstop=Tstop)

    return C



"""
Import functions from derive module
"""

from teneto.derive.derive import derive
from teneto.derive.postprocess import postpro_pipeline
from teneto.derive.postprocess import postpro_fisher
from teneto.derive.postprocess import postpro_standardize
from teneto.derive.postprocess import postpro_boxcox
from teneto.derive.report import gen_report



"""
networkmeasures.bursty_coeff
"""

import numpy as np
from teneto.networkmeasures.intercontacttimes import intercontacttimes


def bursty_coeff(data, calc='edge', nodes='all', subnet=None):
    """
    returns calculates the bursty coefficient. Value > 0
     indicates bursty. Value < 0 periodic/tonic. Value = 0
      implies random.
    As temporalPaths only works with binary undirected edges
     at the moment, weighted edges are assumed to be binary.

    **PARAMETERS**

    :data: This is either:

        :netin: temporal network input (graphlet or contact).

            :nettype: 'bu', 'bd'

        :ICT: dictionary of ICTs (output of *intercontacttimes*).

    :calc: caclulate the bursty coeff over what. Options include

        :'edge': calculate b_coeff on all ICTs between node i and j. (Default)
        :'node': caclulate b_coeff on all ICTs connected to node i.
        :'subnet': calculate b_coeff for each subnetwork (argument subnet then required)

        :'meanEdgePerNode': first calculate the ICTs between node i and j,
         then take the mean over all j.

    :nodes: which do to do. Options include:

        :'all': do for all nodes (default)
        :specify: list of node indexes to calculate.

    :subnet: None (default) or Nx1 vector of subnetwork assignment.
    This returns a "centrality" per subnetwork instead of per node.


    **OUTPUT**

    :b_coeff: bursty coefficienct per (edge or node measure)

        :format: 1d numpy array

    **SEE ALSO**

    intercontacttimes

    **ORIGIN**

    Goh and b_coeffarabasi 2008
    Discrete formulation here from Holme 2012.

    **HISTORY**

    :Modified: Nov 2016, WHT (documentation)
    :Created: Nov 2016, WHT

    """

    if calc == 'subnet' and not subnet:
        raise ValueError("Specified calc='subnet' but no subnet argument provided (list of clusters/modules)")

    ict = 0  # are ict present
    if isinstance(data, dict):
        # This could be done better
        if [k for k in list(data.keys()) if k == 'intercontacttimes'] == ['intercontacttimes']:
            ict = 1
    # if shortest paths are not calculated, calculate them
    if ict == 0:
        data = intercontacttimes(data)

    ict_shape = data['intercontacttimes'].shape

    if len(ict_shape) == 2:
        node_len = ict_shape[0] * ict_shape[1]
    elif len(ict_shape) == 1:
        node_len = 1
    else:
        raise ValueError('more than two dimensions of intercontacttimes')

    if isinstance(nodes, list) and len(ict_shape) > 1:
        node_combinations = [[list(set(nodes))[t], list(set(nodes))[tt]] for t in range(
            0, len(nodes)) for tt in range(0, len(nodes)) if t != tt]
        do_nodes = [np.ravel_multi_index(n, ict_shape) for n in node_combinations]
    else:
        do_nodes = np.arange(0, node_len)

    # Reshae ICTs
    if calc == 'node':
        ict = np.concatenate(data['intercontacttimes'][do_nodes, do_nodes], axis=1)
    elif calc == 'subnet':
        unique_subnet = np.unique(subnet)
        ict_shape = (len(unique_subnet),len(unique_subnet))
        ict = np.array([[None] * ict_shape[0]] * ict_shape[1])
        for i, s1 in enumerate(unique_subnet):
            for j, s2 in enumerate(unique_subnet):
                if s1 == s2:
                    ind = np.triu_indices(sum(subnet==s1),k=1)
                    ict[i,j] = np.concatenate(data['intercontacttimes'][ind[0],ind[1]])
                else:
                    ict[i,j] = np.concatenate(np.concatenate(data['intercontacttimes'][subnet==s1,:][:,subnet==s2]))
        # Quick fix, but could be better
        data['intercontacttimes'] = ict
        do_nodes = np.arange(0,ict_shape[0]*ict_shape[1])

    if len(ict_shape) > 1:
        ict = data['intercontacttimes'].reshape(ict_shape[0] * ict_shape[1])
        b_coeff = np.zeros(len(ict)) * np.nan
    else:
        b_coeff = np.zeros(1) * np.nan
        ict = [data['intercontacttimes']]

    for i in do_nodes:
        if isinstance(ict[i],np.ndarray):
            mu_ict = np.mean(ict[i])
            sigma_ict = np.std(ict[i])
            b_coeff[i] = (sigma_ict - mu_ict) / (sigma_ict + mu_ict)
        else:
            b_coeff[i] = np.nan

    if len(ict_shape) > 1:
        b_coeff = b_coeff.reshape(ict_shape)
    return b_coeff



import numpy as np
from scipy.stats import pearsonr
from scipy.special import betainc


def corrcoef_matrix(matrix):
    # Code originating from http://stackoverflow.com/a/24547964 by http://stackoverflow.com/users/2455058/jingchao

    r = np.corrcoef(matrix)
    rf = r[np.triu_indices(r.shape[0], 1)]
    df = matrix.shape[1] - 2
    ts = rf * rf * (df / (1 - rf * rf))
    pf = betainc(0.5 * df, 0.5, df / (df + ts))
    p = np.zeros(shape=r.shape)
    p[np.triu_indices(p.shape[0], 1)] = pf
    p[np.tril_indices(p.shape[0], -1)] = pf
    p[np.diag_indices(p.shape[0])] = np.ones(p.shape[0])
    return r, p



"""
networkmeasures.reachability_latency
"""

import numpy as np
from teneto.networkmeasures.shortest_temporal_path import shortest_temporal_path


def reachability_latency(data, rratio=1, calc='global'):
    """
    returns global reachability latency.
    This is the r-th longest temporal path. Where r is the number of time If r=1,

    **PARAMETERS**

    :data: This is either:

        :netIn: temporal network input (graphlet or contact).

            :nettype: 'bu'

        :paths: Dictionary of paths (output of shortest_temporal_path).

    :rratio: reachability ratio that the latency is calculated in relation to.
        Value must be over 0 and up to 1.
        1 (default) - all nodes must be reached.
        Other values (e.g. .5 imply that 50% of nodes are reached)
        This is rounded to the nearest node inter.
        E.g. if there are 6 nodes [1,2,3,4,5,6], it will be node 4 (due to round upwards)

    :calc: what to calculate R for:
        :'global': entire network.
        :'nodes': each node.


    **OUTPUT**

    :R: readability latency
        :format: integer (numpy array)

    **SEE ALSO**

    - *temporal_efficiency*
    - *shortest_temporal_path*

    **HISTORY**

    :Modified: Dec 2016, WHT (Documentation)
    :Created: Dec 2016, WHT

    """

    pathdata = 0  # are shortest paths calculated
    if isinstance(data, dict):
        # This could be done better
        if [k for k in list(data.keys()) if k == 'paths'] == ['paths']:
            pathdata = 1
    # if shortest paths are not calculated, calculate them
    if pathdata == 0:
        data = shortest_temporal_path(data)

    netshape = data['paths'].shape

    edges_to_reach = netshape[0] - np.round(netshape[0] * rratio)

    reach_lat = np.zeros([netshape[1], netshape[2]]) * np.nan
    for t_ind in range(0, netshape[2]):
        paths_sort = -np.sort(-data['paths'][:, :, t_ind], axis=1)
        reach_lat[:, t_ind] = paths_sort[:, edges_to_reach]
    if calc == 'global':
        reach_lat = np.nansum(reach_lat)
        reach_lat = reach_lat / ((netshape[0]) * netshape[2])
    elif calc == 'nodes':
        reach_lat = np.nansum(reach_lat, axis=1)
        reach_lat = reach_lat / (netshape[2])
    return reach_lat



"""
Imports when importing networkmeasures
"""

from teneto.networkmeasures.temporal_degree_centrality import temporal_degree_centrality
from teneto.networkmeasures.shortest_temporal_path import shortest_temporal_path
from teneto.networkmeasures.temporal_closeness_centrality import temporal_closeness_centrality
from teneto.networkmeasures.intercontacttimes import intercontacttimes
from teneto.networkmeasures.volatility import volatility
from teneto.networkmeasures.bursty_coeff import bursty_coeff
from teneto.networkmeasures.fluctuability import fluctuability
from teneto.networkmeasures.temporal_efficiency import temporal_efficiency
from teneto.networkmeasures.reachability_latency import reachability_latency
from teneto.networkmeasures.sid import sid



import teneto.utils as utils
import numpy as np


def volatility(net, distance_func_name='default', calc='global', subnet=None):
    """
    volatility of temporal networks. This is the average distance between consecutive time points of graphlets (difference is caclualted either globally, per edge)

    **PARAMETERS**

    :net: temporal network input (graphlet or contact)

        :net nettype: 'bu','bd','wu','wd'

    :D: distance function. Following options available: 'default', 'hamming', 'euclidean'. (Default implies hamming for binary networks, euclidean for weighted).
    :calc: version of volaitility to caclulate. Possibilities include:

        :'global': (default): the average distance of all nodes for each consecutive time point).
        :'edge': average distance between consecutive time points for each edge). Takes considerably longer
        :'node': (i.e. returns the average per node output when calculating volatility per 'edge').
        :'time': returns volatility per time point
        :'subnet': returns volatility per subnetwork id (see subnet). Also is returned per time-point and this may be changed in the future (with additional options)
    :'subnet': vector of integers.
    
    :Note: Index of subnetworks are returned "as is" with a shape of [max(subnet)+1,max(subnet)+1]. So if the indexes used are [1,2,3,5], V.shape==(6,6). The returning V[1,2] will correspond indexes 1 and 2. And missing index (e.g. here 0 and 4 will be NANs in rows and columns). If this behaviour is unwanted, call clean_subnetdexes first. This will probably change.

    :network:

    **OUTPUT**

    :vol: Volatility

        :format: scalar (calc='global'),
            1d numpy array (calc='node'),
            2d numpy array (calc='edge')

    **SEE ALSO**
    - *utils.getDistanceFunction*

    **HISTORY**
    Modified - Aug 2016, WHT (PEP8, cleanup)
    Modified - Dec 2016, WHT (documentation, cleanup)
    Created - Dec 2016, WHT
    """

    # Get input (C or G)
    net, netinfo = utils.process_input(net, ['C', 'G', 'TO'])

    distance_func_name = utils.check_distance_funciton_input(
        distance_func_name, netinfo)

    if not isinstance(distance_func_name, str):
        raise ValueError('Distance metric must be a string')

    # If not directional, only calc on the uppertriangle
    if netinfo['nettype'][1] == 'd':
        ind = np.triu_indices(net.shape[0], k=-net.shape[0])
    elif netinfo['nettype'][1] == 'u':
        ind = np.triu_indices(net.shape[0], k=1)

    if calc == 'subnet':
        # Make sure subnet is np array for indexing later on.
        subnet = np.array(subnet)
        if len(subnet) != netinfo['netshape'][0]:
            raise ValueError(
                'When processing per network, subnet vector must equal the number of nodes')
        if subnet.min() < 0:
            raise ValueError(
                'Subnetwork assignments must be positive integers')

    # Get chosen distance metric fucntion
    distance_func = utils.getDistanceFunction(distance_func_name)

    if calc == 'global':
        vol = np.mean([distance_func(net[ind[0], ind[1], t], net[ind[0],
                                                                       ind[1], t + 1]) for t in range(0, net.shape[-1] - 1)])
    elif calc == 'time':
        vol = [distance_func(net[ind[0], ind[1], t], net[ind[0], ind[1], t + 1])
               for t in range(0, net.shape[-1] - 1)]
    # This takes quite a bit of time to loop through. When calculating per edge/node.
    elif calc == 'edge' or calc == 'node':
        vol = np.zeros([net.shape[0], net.shape[1]])
        for i in ind[0]:
            for j in ind[1]:
                vol[i, j] = np.mean([distance_func(
                    net[i, j, t], net[i, j, t + 1]) for t in range(0, net.shape[-1] - 1)])
        if netinfo['nettype'][1] == 'u':
            vol = vol + np.transpose(vol)
        if calc == 'node':
            vol = np.sum(vol, axis=1)
    elif calc == 'subnet':
        net_id = set(subnet)
        vol = np.zeros([max(net_id) + 1, max(net_id) +
                        1, netinfo['netshape'][-1] - 1])
        for net1 in net_id:
            for net2 in net_id:
                vol[net1, net2, :] = [distance_func(net[subnet == net1][:, subnet == net2, t].flatten(),
                                                    net[subnet == net1][:, subnet == net2, t + 1].flatten()) for t in range(0, net.shape[-1] - 1)]
    elif calc == 'withinsubnet':
        within_ind = np.array([[ind[0][n], ind[1][n]] for n in range(
            0, len(ind[0])) if subnet[ind[0][n]] == subnet[ind[1][n]]])
        vol = [distance_func(net[within_ind[:, 0], within_ind[:, 1], t], net[within_ind[:, 0],
                                                                                   within_ind[:, 1], t + 1]) for t in range(0, net.shape[-1] - 1)]
    elif calc == 'betweensubnet':
        between_ind = np.array([[ind[0][n], ind[1][n]] for n in range(
            0, len(ind[0])) if subnet[ind[0][n]] != subnet[ind[1][n]]])
        vol = [distance_func(net[between_ind[:, 0], between_ind[:, 1], t], net[between_ind[:,
                                                                                                 0], between_ind[:, 1], t + 1]) for t in range(0, net.shape[-1] - 1)]

    return vol





def correct_pvalues_for_multiple_testing(pvalues, correction_type="Benjamini-Hochberg"):
    """

    Code originates at http://stackoverflow.com/a/21739593 by http://stackoverflow.com/users/955278/emre

    Modfied for python3, Jan17

    """

    from numpy import array, empty
    pvalues = array(pvalues)
    n = float(pvalues.shape[0])
    new_pvalues = empty(int(n))
    if correction_type == "Bonferroni":
        new_pvalues = n * pvalues
    elif correction_type == "Bonferroni-Holm":
        values = [(pvalue, i) for i, pvalue in enumerate(pvalues)]
        values.sort()
        for rank, vals in enumerate(values):
            pvalue, i = vals
            new_pvalues[i] = (n - rank) * pvalue
    elif correction_type == "Benjamini-Hochberg":
        values = [(pvalue, i) for i, pvalue in enumerate(pvalues)]
        values.sort()
        values.reverse()
        new_values = []
        for i, vals in enumerate(values):
            rank = n - i
            pvalue, index = vals
            new_values.append((n / rank) * pvalue)
        for i in range(0, int(n) - 1):
            if new_values[i] < new_values[i + 1]:
                new_values[i + 1] = new_values[i]
        for i, vals in enumerate(values):
            pvalue, index = vals
            new_pvalues[index] = new_values[i]
    return new_pvalues



import numpy as np

def acceptance_probabailities(T,costCurrent,costNew,costGlobal,cCurrent,cNew,cGlobal):
    if costCurrent >= costNew:
        costCurrent = np.array(costNew)
        cCurrent = np.array(cNew)
    else:
        p=np.exp(-(costNew-costCurrent)/T)
        if p>=np.random.rand():
            costCurrent = np.array(costNew)
            cCurrent=np.array(cNew)
    #This part is specified if the algorithem explores too far and goes away from already obtained optimal solution
    if costCurrent<costGlobal:
        costGlobal = np.array(costCurrent)
        cGlobal = np.array(cCurrent)
    return costCurrent,costGlobal,cCurrent,cGlobal


def simulated_annealing(netIn,costFun,costFunParams,Cnprior=1,repeat=1,cooldown = 0.995, f=0.5,Tstop=10e-100):

    """

    Simulated annelaing. Still in testing phase (see NOTE).

    **PARAMETERS**

    :netIn: 2D network input
    :costFun: cost function that takes input costFun(netIn,clusterAssignment,costFunParams)
    :costFunParams: dictionary of parameters for costFun
    :Cnprior: Prior on number of clusters believes to exists.
    :repeat: how many times to repeat
    :cooldown: reduction in temeprature for each itteration
    :f: proportion of nodes shuffled each iteration (also how many merge/splitting attempts per round)
    :Tstop: terminate optimization. Increase number for quicker performance.

    **OUTPUT**

    :cCurrent: Cluster assignment of network (glaobal minimum found)
    :cost: cost function output

    **NOTE**

    1. Currently a bit slow due to their being no stopping condition and low Tstop (I reasoned better safe than sorry)
    2. For merging/splitting new networks (found a bit ambigious at source), this was done at a seperate step than shuffling node clster membership (recaclulating Q). Nodes were only shuffled once per round of T (and the number of nodes shuffled is dictated by F). Cluster merging/splitting attempts are tried multiple times per T (if allowed by 0)
    For each attempt at merging/splitting, Q is recalculated.


    :REFERENCE:
    *Guimerà and Amaral (2005), Nature*.

    :HISTORY:

    Created, Jan17, WHT

    """

    n_i = int(np.round(f*netIn.shape[0]))
    n_c = int(np.round(f*netIn.shape[0]))

    costGlobal = 1000000
    cGlobal = []

    for r in range(0,repeat):

        #Initial state
        Cn = Cnprior
        cCurrent = np.random.randint(0,Cn,netIn.shape[0])
        costCurrent = costFun(netIn,cCurrent,costFunParams)
        T=1

        while T>Tstop:
            #
            #costNew = costFun(netIn,cNew)

            # Get Neighbours (Shuffle nodes across clusters)
            update = np.random.randint(0,netIn.shape[0],n_i)
            cNew = np.array(cCurrent)
            cNew[update] = np.random.randint(0,Cn,n_i)

            # Cost function of new cluster
            costNew = costFun(netIn,cNew,costFunParams)

            # Update
            costCurrent,costGlobal,cCurrent,cGlobal=acceptance_probabailities(T,costCurrent,costNew,costGlobal,cCurrent,cNew,cGlobal)


            # get neighbours part 2
            # Add or split clusters (slightly unclear in Guimerà and Amaral how they achieved this and I asusme equal probabalistic to split, merge)
            for splitmerge in range(0,n_c):
                cNew = np.array(cCurrent)
                if Cn > 1 and Cn < netIn.shape[0]:
                    r=np.random.rand()
                    if r>=.5:
                        merge = 1
                        split = 0
                    else:
                        merge = 0
                        split = 1
                elif Cn == netIn.shape[0]:
                    merge = 1
                    split = 0
                else:
                    merge = 0
                    split = 1
                # If f is high and too many merge and splittings occur so set to
                if split == 1:
                    if split > Cn:
                        split = Cn
                    coi_split= np.random.randint(0,Cn)
                    roi=np.where(cNew==coi_split)[0]
                    update = np.random.randint(0,2,len(roi))
                    if len(np.unique(update))>1:
                        cNew[roi[update==1]]=Cn
                        Cn += 1
                if merge == 1:
                    if merge > Cn:
                        merge = Cn
                    coi_merge=np.argsort(np.random.rand(Cn))
                    roi1=np.where(cNew==coi_merge[1])[0]
                    cNew[roi1]=coi_merge[0]
                    #Set max to empty cluster index
                    cNew[cNew==cNew.max()]=coi_merge[1]
                    Cn -= 1

                costNew = costFun(netIn,cNew,costFunParams)

                costCurrent,costGlobal,cCurrent,cGlobal=acceptance_probabailities(T,costCurrent,costNew,costGlobal,cCurrent,cNew,cGlobal)

            T=T*cooldown

    return cGlobal, costGlobal



import matplotlib.pyplot as plt
import numpy as np
from teneto.utils import contact2graphlet, checkInput

from scipy import ndimage

plt.rcParams['axes.facecolor'] = 'white'


def graphlet_stack_plot(netIn, ax, q=10, cmap='Reds', gridcolor='k', borderwidth=2, bordercolor=[0, 0, 0], Fs=1, timeunit='', t0=1, sharpen='yes', vminmax='minmax'):
    '''
    Returns matplotlib axis handle for graphlet_stack_plot. This is a row of transformed connectivity matrices to look like a 3D stack.
    **PARAMETERS**
    :netIn: network input (graphlet or contact)
    :ax: matplotlib ax handles.
    :q: quality. Increaseing this will lead to smoother axis but take up more memory.
    :cmap: colormap (matplotlib) of graphlets
    :Fs: sampling rate. Same as contact-representation (if netIn is contact, and input is unset, contact dictionary is used)
    :timeunit: for plotting. Same as contact-representation (if netIn is contact, and input is unset, contact dictionary is used)
    :t0: what should the first time point be called. Should be integer. Default 1.
    :gridcolor: The color of the grid section of the graphlets. Set to 'none' if not wanted.
    :borderwidth: Integer that scales the size of border. (at the moment it cannot be set to 0.)
    :bordorcolor: color of the border (at the moment it must be in RGB values between 0 and 1 -> this will be changed sometime in the future)
    :vminmax: 'maxabs', 'minmax' (default), or list/array with length of 2. Specifies the min and max colormap value of graphlets. Maxabs entails [-max(abs(G)),max(abs(G))], minmax entails [min(G), max(G)].
    **OUTPUT**
    :ax: matplotlib ax handle
    **NOTE**
    This function can require a lot of RAM with larger networks.
    At the momenet bordercolor cannot be set to zero. To remove border, set bordorwidth=1 and bordercolor=[1,1,1] for temporay workaround.
    **SEE ALSO**
    - *circle_plot*
    - *slice_plot*
    **HISTORY**
    :Created: Dec 2016, WHT
    '''

    # Get input type (C, G, TO)
    inputType = checkInput(netIn)

    # Convert TO to C representation
    if inputType == 'TO':
        netIn = netIn.contact
        inputType = 'C'
    # Convert C representation to G
    if inputType == 'C':
        nettype = netIn['nettype']
        if timeunit == '':
            timeunit = netIn['timeunit']
        if t0 == 1:
            t0 = netIn['t0']
        if Fs == 1:
            Fs = netIn['Fs']
        netIn = contact2graphlet(netIn)

    if timeunit != '':
        timeunit = ' (' + timeunit + ')'

    if not isinstance(borderwidth, int):
        borderwidth = int(borderwidth)
        print('Warning: borderwidth should be an integer. Converting to integer.')

    # x and y ranges for each of the graphlet plots
    v = np.arange(0, netIn.shape[0] + 1)
    vr = np.arange(netIn.shape[0], -1, -1)
    # Preallocatie matrix

    if vminmax == '' or vminmax == 'absmax' or vminmax == 'maxabs':
        vminmax = [-np.nanmax(np.abs(netIn)), np.nanmax(np.abs(netIn))]
    elif vminmax == 'minmax':
        vminmax = [np.nanmin(netIn), np.nanmax(netIn)]

    qb = q * borderwidth
    figmat = np.zeros([80 * q + (qb * 2), int(((netIn.shape[-1]) *
                                               (80 * q) + (qb * 2)) - ((netIn.shape[-1] - 1) * q * 80) / 2), 4])
    for n in range(0, netIn.shape[-1]):
        # Create graphlet
        figtmp, axtmp = plt.subplots(1, facecolor='white', figsize=(q, q), dpi=80)
        axtmp.pcolormesh(v, vr, netIn[:, :, n], cmap=cmap, edgecolor=gridcolor,
                         linewidth=q * 2, vmin=vminmax[0], vmax=vminmax[1])
        axtmp.set_xticklabels('')
        axtmp.set_yticklabels('')
        axtmp.set_xticks([])
        axtmp.set_yticks([])
        x0, x1 = axtmp.get_xlim()
        y0, y1 = axtmp.get_ylim()
        axtmp.set_aspect((x1 - x0) / (y1 - y0))
        axtmp.spines['left'].set_visible(False)
        axtmp.spines['right'].set_visible(False)
        axtmp.spines['top'].set_visible(False)
        axtmp.spines['bottom'].set_visible(False)
        plt.subplots_adjust(left=0, bottom=0, right=1,
                            top=1, wspace=0, hspace=0)

        # Convert graphlet to RGB values
        figtmp.canvas.draw()
        figmattmp = np.fromstring(
            figtmp.canvas.tostring_rgb(), dtype=np.uint8, sep='')
        figmattmp = figmattmp.reshape(
            figtmp.canvas.get_width_height()[::-1] + (3,))

        # Close figure for memory
        plt.close(figtmp)

        # Manually add a border

        figmattmp_withborder = np.zeros(
            [figmattmp.shape[0] + (qb * 2), figmattmp.shape[1] + (qb * 2), 3]) + (np.array(bordercolor) * 255)
        figmattmp_withborder[qb:-qb, qb:-qb, :] = figmattmp

        # Make corners rounded. First make a circle and then take the relevant quarter for each corner.
        y, x = np.ogrid[-qb: qb + 1, -qb: qb + 1]
        mask = x * x + y * y <= qb * qb
        # A little clumsy. Should improve
        Mq1 = np.vstack([[mask[:qb, :qb] == 0], [mask[:qb, :qb] == 0], [
                        mask[:qb, :qb] == 0]]).transpose([1, 2, 0])
        figmattmp_withborder[:qb, :qb, :][Mq1] = 255
        Mq1 = np.vstack([[mask[:qb, -qb:] == 0], [mask[:qb, -qb:]
                                                  == 0], [mask[:qb, -qb:] == 0]]).transpose([1, 2, 0])
        figmattmp_withborder[:qb, -qb:, :][Mq1] = 255
        Mq1 = np.vstack([[mask[-qb:, :qb] == 0], [mask[-qb:, :qb]
                                                  == 0], [mask[-qb:, :qb] == 0]]).transpose([1, 2, 0])
        figmattmp_withborder[-qb:, :qb, :][Mq1] = 255
        Mq1 = np.vstack([[mask[-qb:, -qb:] == 0], [mask[-qb:, -qb:]
                                                   == 0], [mask[-qb:, -qb:] == 0]]).transpose([1, 2, 0])
        figmattmp_withborder[-qb:, -qb:, :][Mq1] = 255

        #scale and sheer
        scale = np.matrix([[1.5, 0, 0], [0, 3, 0], [0, 0, 1]])
        sheer = np.matrix([[1, np.tan(np.pi / 12), 0], [0, 1, 0], [0, 0, 1]])

        # apply affine transformation
        figmattmp = ndimage.affine_transform(
            figmattmp_withborder, sheer * (scale), offset=[-35 * q, 0, 0], cval=255)

        # At the moment the alpha part does not work if the background colour is anything but white.
        # Also used for detecting where the graphlets are in the image.
        trans = np.where(np.sum(figmattmp, axis=2) == 255 * 3)
        alphamat = np.ones([figmattmp.shape[0], figmattmp.shape[0]])
        alphamat[trans[0], trans[1]] = 0
        figmattmp = np.dstack([figmattmp, alphamat])

        # Add graphlet to matrix
        if n == 0:
            figmat[:, n * (80 * q):((n + 1) * (80 * q) + (qb * 2))] = figmattmp
        else:
            figmat[:, n * (80 * q) - int((n * q * 80) / 2):int(((n + 1)
                                                                * (80 * q) + (qb * 2)) - (n * q * 80) / 2)] = figmattmp

    # Fix colours - due to imshows weirdness when taking nxnx3
    figmat[:, :, 0:3] = figmat[:, :, 0:3] / 255
    # Cut end of matrix off that isn't need
    figmat = figmat[:, :-int((q / 2) * 80), :]
    fid = np.where(figmat[:, :, -1] > 0)
    fargmin = np.argmin(fid[0])
    ymax = np.max(fid[0])
    yright = np.max(np.where(figmat[:, fid[1][fargmin], -1] > 0))
    xtickloc = np.where(figmat[ymax, :, -1] > 0)[0]
    # In case there are multiple cases of xtickloc in same graphlet (i.e. they all have the same lowest value)
    xtickloc = np.delete(xtickloc, np.where(np.diff(xtickloc) == 1)[0] + 1)

    fid = np.where(figmat[:, :, -1] > 0)
    ymin = np.min(fid[0])
    topfig = np.where(figmat[ymin, :, -1] > 0)[0]
    topfig = topfig[0:len(topfig):int(len(topfig) / netIn.shape[-1])]

    # Make squares of non transparency around each figure (this fixes transparency issues when white is in the colormap)
    # for n in range(0,len(topfig)):
    # fid=np.where(figmat[ymin:ymax,xtickloc[n]:topfig[n],-1]==0)
    # figmat[ymin:ymax,xtickloc[n]:topfig[n],:3][fid[0],fid[1]]=1
    # figmat[ymin+q:ymax-q,xtickloc[n]+q:topfig[n]-q,-1]=1

    # Create figure
    # Sharped edges of figure with median filter
    if sharpen == 'yes':
        figmat[:, :, :-1] = ndimage.median_filter(figmat[:, :, :-1], 3)
    ax.imshow(figmat[:, :, :-1], zorder=1)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.set_xticklabels('')
    ax.set_yticklabels('')
    ax.set_xticks([])
    ax.set_yticks([])

    L = int((((netIn.shape[-1] - 3) + 1) * (80 * q) +
             (qb * 2)) - ((netIn.shape[-1] - 3) * q * 80) / 2 - q)
    [ax.plot(range(topfig[i], xt), np.zeros(len(range(topfig[i], xt))) + yright,
             color='k', linestyle=':', zorder=2) for i, xt in enumerate(xtickloc[1:])]
    ax.plot(range(0, L), np.zeros(L) + ymax,
            color='k', linestyle=':', zorder=2)
    [ax.plot(np.zeros(q * 10) + xt, np.arange(ymax, ymax + q * 10),
             color='k', linestyle=':', zorder=2) for xt in xtickloc]
    [ax.text(xt, ymax + q * 20, str(round((i + t0) * Fs, 5)),
             horizontalalignment='center',) for i, xt in enumerate(xtickloc)]

    ylim = ax.axes.get_ylim()
    xlim = ax.axes.get_xlim()
    ax.set_ylim(ylim[0] + q * 15, 0)
    ax.set_xlim(xlim[0] - q * 20, xlim[1])
    ax.set_xlabel('Time' + timeunit)
    return ax



import itertools
import teneto
import os
import re
from bids.grabbids import BIDSLayout
import numpy as np
import inspect
import pandas as pd
import statsmodels.formula.api as smf
import seaborn as sns
import matplotlib.pyplot as plt
import pickle
import traceback

#class NetworkMeasures:
#    def __init__(self,**kwargs):
#        pass

    #def temporal_degree_centrality(self,**kwargs):
    #    print(self)
    #    print(teneto.networkmeasures.temporal_degree_centrality(self,**kwargs))

class TenetoBIDS:

    #networkmeasures = NetworkMeasures(self)

    def __init__(self, BIDS_dir, pipeline=None, pipeline_subdir=None, parcellation=None, space=None, subjects='all', sessions='all', runs='all', tasks='all', last_analysis_step=None, analysis_steps=None, bad_subjects=None, confound_pipeline=None, raw_data_exists=True):
        """
        Parameters
        ----------

        BIDS_dir : str
            string to BIDS directory
        pipeline : str
            the directory that is in the BIDS_dir/derivatives/<pipeline>/
        pipeline_subdir : str, optional
            the directory that is in the BIDS_dir/derivatives/<pipeline>/sub-<subjectnr/func/ses-<sesnr>/<pipeline_subdir>
        parcellation : str, optional
            parcellation name
        space : str, optional
            different nomralized spaces
        subjects : str or list, optional
            can be part of the BIDS file name
        sessions : str or list, optional
            can be part of the BIDS file name
        runs : str or list, optional
            can be part of the BIDS file name
        tasks : str or list, optional
            can be part of the BIDS file name
        analysis_steps : str or list, optional
            any tags that exist in the filename (e.g. 'bold' or 'preproc')
        bad_subjects : list or str, optional
            Removes these subjects from the analysis
        confound_pipeline : str, optional
            If the confounds file is in another derivatives directory than the pipeline directory, set it here.
        raw_data_exists : bool, optional
            Default is True. If the unpreprocessed data is not present in BIDS_dir, set to False. Some BIDS funcitonality will be lost.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)

        self.contact = []
        if raw_data_exists:
            self.BIDS = BIDSLayout(BIDS_dir)
        else:
            self.BIDS = 'Raw data was flagged as not present in directory structure.'
        self.BIDS_dir = BIDS_dir
        self.pipeline = pipeline
        self.confound_pipeline = confound_pipeline
        self.raw_data_exists = raw_data_exists
        if not pipeline_subdir:
            self.pipeline_subdir = ''
        else:
            self.pipeline_subdir = pipeline_subdir
        self.parcellation = parcellation
        self.space = space
        if self.BIDS_dir[-1] != '/':
            self.BIDS_dir = self.BIDS_dir + '/'
        if subjects == 'all':
            if self.raw_data_exists:
                self.subjects = self.BIDS.get_subjects()
            else:
                self.set_subjects()
        if sessions == 'all' and self.raw_data_exists:
            self.sessions = self.BIDS.get_sessions()
        elif self.raw_data_exists:
            self.set_sessions(sessions)
        else:
            self.sessions = []
        if tasks == 'all' and self.raw_data_exists:
            self.tasks = self.BIDS.get_tasks()
        elif self.raw_data_exists:
            self.set_tasks(tasks)
        else:
            self.tasks = []
        if runs == 'all' and self.raw_data_exists:
            self.runs = self.BIDS.get_runs()
        elif self.raw_data_exists:
            self.set_runs(runs)
        else:
            self.runs = []
        if not last_analysis_step:
            self.last_analysis_step = ''
        else:
            self.last_analysis_step = last_analysis_step
        if isinstance(analysis_steps,str):
            self.analysis_steps = [analysis_steps]
        elif isinstance(analysis_steps,list):
            self.analysis_steps = analysis_steps
        else:
            self.analysis_steps = ''

        if bad_subjects == None:
            self.bad_subjects = None
        else:
            self.set_bad_subjects(bad_subjects)


    def add_history(self, fname, fargs, init=0):
        """
        Adds a processing step to TenetoBIDS.history.
        """
        if init == 1:
            self.history = []
        self.history.append([fname,fargs])


    def make_functional_connectivity(self):
        """
        Makes connectivity matrix for each of the subjects.

        Saves data in derivatives/teneto_<version>/.../fc/
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        files = self.get_selected_files(quiet=1)

        R_group = []
        for i, f in enumerate(files):

            # ADD MORE HERE (csv, json, nifti)
            if f.split('.')[-1] == 'npy':
                data = np.load(f)
            else:
                raise ValueError('derive can only load npy files at the moment')

            file_name = f.split('/')[-1].split('.')[0]
            save_name = file_name + '_fc'
            paths_post_pipeline = f.split(self.pipeline)

            if self.pipeline_subdir:
                paths_post_pipeline = paths_post_pipeline[1].split(self.pipeline_subdir)[0]
            else:
                paths_post_pipeline = paths_post_pipeline[1].split(file_name)[0]
            save_dir = self.BIDS_dir + '/derivatives/' + 'teneto_' + teneto.__version__ + '/' + paths_post_pipeline + '/fc/'

            if not os.path.exists(save_dir):
                os.makedirs(save_dir)

            R = teneto.misc.corrcoef_matrix(data)[0]
            # Fisher transform of subject R values before group average
            R_group.append(np.arctan(R))
            np.save(save_dir + save_name + '.npy', R)



    def derive(self, params, update_pipeline=True, tag=None):

        """
        Derive time-varying connectivity on the selected files.

        params : dict.
            See teneto.derive.derive for the structure of the param dictionary.

        update_pipeline : bool
            If true, the object updates the selected files with those derived here.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)

        files = self.get_selected_files(quiet=1)
        confound_files = self.get_confound_files(quiet=1)
        if confound_files:
            confounds_exist = True

        if not tag:
            tag = ''
        else:
            tag = '_' + tag

        for i, f in enumerate(files):

            # ADD MORE HERE (csv, json, nifti)
            if f.split('.')[-1] == 'npy':
                data = np.load(f)
            else:
                raise ValueError('derive can only load npy files at the moment')

            file_name = f.split('/')[-1].split('.')[0]
            save_name = file_name + '_tvcmethod-' + params['method'] + tag + '_tvc'
            paths_post_pipeline = f.split(self.pipeline)

            if self.pipeline_subdir:
                paths_post_pipeline = paths_post_pipeline[1].split(self.pipeline_subdir)[0]
            else:
                paths_post_pipeline = paths_post_pipeline[1].split(file_name)[0]
            save_dir = self.BIDS_dir + '/derivatives/' + 'teneto_' + teneto.__version__ + '/' + paths_post_pipeline + '/tvc/'

            if 'weight-var' in params.keys():
                if params['weight-var'] == 'from-subject-fc':
                    fc_dir = self.BIDS_dir + '/derivatives/' + 'teneto_' + teneto.__version__ + '/' + paths_post_pipeline + '/fc/'
                    f = os.listdir(fc_dir)
                    params['weight-var'] = np.load(fc_dir + f[0])

            if 'weight-mean' in params.keys():
                if params['weight-mean'] == 'from-subject-fc':
                    fc_dir = self.BIDS_dir + '/derivatives/' + 'teneto_' + teneto.__version__ + '/' + paths_post_pipeline + '/fc/'
                    f = os.listdir(fc_dir)
                    params['weight-mean'] = np.load(fc_dir + f[0])

            params['report'] = 'yes'
            params['report_path'] =  self.BIDS_dir + '/derivatives/' + 'teneto_' + teneto.__version__ + '/' + paths_post_pipeline + '/tvc/report/'
            params['report_filename'] =  save_name + '_derivationreport.html'

            if not os.path.exists(save_dir):
                os.makedirs(save_dir)

            if not os.path.exists(params['report_path']):
                os.makedirs(params['report_path'])

            dfc = teneto.derive.derive(data,params)

            np.save(save_dir + save_name + '.npy', dfc)

            if confounds_exist:
                analysis_step = 'tvc-derive'
                if confound_files[i].split('.')[-1] == 'csv':
                    delimiter = ','
                elif confound_files[i].split('.')[-1] == 'tsv':
                    delimiter = '\t'
                df = pd.read_csv(confound_files[i],sep=delimiter)
                df = df.fillna(df.median())
                ind = np.triu_indices(dfc.shape[0], k=1)
                dfc_df = pd.DataFrame(dfc[ind[0],ind[1],:].transpose())
                #NOW CORRELATE DF WITH DFC BUT DFC INDEX NOT DF.
                dfc_df_z = (dfc_df - dfc_df.mean())
                df_z = (df - df.mean())
                R_df = dfc_df_z.T.dot(df_z).div(len(dfc_df)).div(df_z.std(ddof=0)).div(dfc_df_z.std(ddof=0), axis=0)
                R_df_describe = R_df.describe()
                desc_index = R_df_describe.index
                confound_report_dir = params['report_path'] + '/' + analysis_step + '_vs_confounds/'
                confound_report_figdir = confound_report_dir + 'figures/'
                if not os.path.exists(confound_report_figdir):
                    os.makedirs(confound_report_figdir)
                report = '<html><body>'
                report += '<h1> Correlation of ' + analysis_step + ' and confounds.</h1>'
                for c in R_df.columns:
                    fig,ax = plt.subplots(1)
                    ax = sns.distplot(R_df[c],hist=False, color='m', ax=ax, kde_kws={"shade": True})
                    fig.savefig(confound_report_figdir + c + '.png')
                    plt.close(fig)
                    report += '<h2>' + c + '</h2>'
                    for ind_name,r in enumerate(R_df_describe[c]):
                        report += str(desc_index[ind_name]) + ': '
                        report += str(r) + '<br>'
                    report += 'Distribution of corrlation values:'
                    report += '<img src=' + confound_report_figdir + c + '.png><br><br>'
                report += '</body></html>'

            with open(confound_report_dir + analysis_step + '_vs_confounds.html', 'w') as file:
                file.write(report)

            file.close()

        if update_pipeline == True:
            if not self.confound_pipeline and len(self.get_confound_files(quiet=1)) > 0:
                self.set_confound_pipeline = self.pipeline
            self.set_pipeline('teneto_' + teneto.__version__)
            self.set_pipeline_subdir('tvc')
            self.set_last_analysis_step('tvc')
            if tag:
                self.analysis_steps += [tag[1:]]


    def networkmeasures(self, measure=None, measure_params={}, load_tag=None, save_tag=None):
        """
        Runs a network measure

        For available funcitons see: teneto.networkmeasures

        Parameters
        ----------

        measure : str or list
            Mame of function(s) from teneto.networkmeasures that will be run.

        measure_params : dict or list of dctionaries)
            Containing kwargs for the argument in measure.

        tag : str
            Add additional tag to filenames.

        Note
        ----
        If self.network_communities exist, subnet=True can be specified for subnet options instead of supplying the network atlas.

        Returns
        -------

        Saves in ./BIDS_dir/derivatives/teneto/sub-NAME/func/tvc/temporal-network-measures/MEASURE/
        Load the measure with tenetoBIDS.load_network_measure
        """

        self.add_history(inspect.stack()[0][3], locals(), 1)

        module_dict = inspect.getmembers(teneto.networkmeasures)
        # Remove all functions starting with __
        module_dict={m[0]:m[1] for m in module_dict if m[0][0:2]!='__'}
        # measure can be string or list
        if isinstance(measure, str):
            measure = [measure]
        # measure_params can be dictionaary or list of dictionaries
        if     isinstance(measure_params, dict):
            measure_params = [measure_params]
        if measure_params and len(measure) != len(measure_params):
            raise ValueError('Number of identified measure_params (' + str(len(measure_params)) + ') differs from number of identified measures (' + str(len(measure)) + '). Leave black dictionary if default methods are wanted')


        # Check that specified measure is valid.
        flag = [n for n in measure if n not in module_dict.keys()]
        if flag:
            print('Specified measure(s): ' + ', '.join(flag) + ' not valid.')
        if not measure or flag:
            print('Following measures are valid (specified as string or list): \n - ' + '\n - '.join(module_dict.keys()))

        files = self.get_selected_files(quiet=1)


        if not load_tag:
            load_tag = ''

        if not save_tag:
            save_tag = ''
        else:
            save_tag = '_' + save_tag

        for f in files:

            if load_tag in f:

                # ADD MORE HERE (csv, json, nifti)
                if f.split('.')[-1] == 'npy':
                    data = np.load(f)
                else:
                    raise ValueError('derive can only load npy files at the moment')

                save_dir_base = '/'.join(f.split('/')[:-1]) + '/temporal-network-measures/'

                file_name = f.split('/')[-1].split('.')[0]

                for i, m in enumerate(measure):

                    # The following 12 lines get the dimord
                    if 'calc' in measure_params[i]:
                        c = measure_params[i]['calc']
                        cs = '_calc-' + c
                    else:
                        c = ''
                        cs = ''
                    if 'subnet' in measure_params[i]:
                        s = 'subnet'
                    else:
                        s = ''
                    dimord = teneto.utils.get_dimord(m,c,s)
                    dimord_str = ''
                    if dimord != 'unknown':
                        dimord_str = '_dimord-' + dimord

                    if 'subnet' in measure_params[i]:
                        if measure_params[i]['subnet'] == True:
                            measure_params[i]['subnet'] = list(self.network_communities_['network_id'].values)

                    sname = m.replace('_','-')
                    if not os.path.exists(save_dir_base + sname):
                        os.makedirs(save_dir_base + sname)

                    save_name = file_name + '_' + sname + cs + dimord_str + save_tag
                    netmeasure = module_dict[m](data,**measure_params[i])

                    np.save(save_dir_base + sname + '/' + save_name, netmeasure)

    def get_space_alternatives(self,quiet=0):
        """
        Returns which space alternatives can be identified in the BIDS derivatives structure. Spaces are denoted with the prefix "space-".
        """
        if not self.pipeline:
            print('Please set pipeline first.')
            self.get_pipeline_alternatives()
        else:
            space_alternatives = []
            if self.sessions:
                ses = '/ses-' + self.sessions + '/'
            else:
                ses = ''
            for s in self.BIDS.get_subjects():
                derdir_files = os.listdir(self.BIDS_dir + '/derivatives/' + self.pipeline + '/' + self.pipeline_subdir +'/sub-' + s + '/' + ses + 'func/')
                space_alternatives += [re.split('[_.]',f.split('_space-')[1])[0] for f in derdir_files if re.search('_space-',f)]
            space_alternatives = set(space_alternatives)
            if quiet == 0:
                print('Space alternatives: ' + ', '.join(space_alternatives))
            return list(space_alternatives)

    def get_pipeline_alternatives(self,quiet=0):
        """
        The pipeline are the different outputs that are placed in the ./derivatives directory.

        get_pipeline_alternatives gets those which are found in the specified BIDS directory structure.
        """
        if not os.path.exists(self.BIDS_dir + '/derivatives/'):
            print('Derivative directory not found. Is the data preprocessed?')
        else:
            pipeline_alternatives = os.listdir(self.BIDS_dir + '/derivatives/')
            if quiet == 0:
                print('Derivative alternatives: ' + ', '.join(pipeline_alternatives))
            return list(pipeline_alternatives)

    def get_pipeline_subdir_alternatives(self,quiet=0):
        """
        Note
        -----

        This function currently returns the wrong folders and will be fixed in the future.

        This function should return ./derivatives/pipeline/sub-xx/[ses-yy/][func/]/pipeline_subdir
        But it does not care about ses-yy at the moment.
        """
        if not self.pipeline:
            print('Please set pipeline first.')
            self.get_pipeline_alternatives()
        else:
            pipeline_subdir_alternatives = []
            for s in self.BIDS.get_subjects():
                derdir_files = os.listdir(self.BIDS_dir + '/derivatives/' + self.pipeline + '/' + s + '/')
                pipeline_subdir_alternatives += [f for f in derdir_files if os.path.isdir(f)]
            pipeline_subdir_alternatives = set(pipeline_subdir_alternatives)
            if quiet == 0:
                print('Pipeline_subdir alternatives: ' + ', '.join(pipeline_subdir_alternatives))
            return list(pipeline_subdir_alternatives)

    def get_selected_files(self,quiet=0):
        """
        Parameters
        ----------
        quiet: int
            If 1, prints results. If 0, no results printed.

        Returns
        -------
        found_files : list
            The files which are currently selected with the current using the set pipeline, pipeline_subdir, space, parcellation, tasks, runs, subjects etc. There are the files that will generally be used if calling a make_ function.
        """
        # This could be mnade better
        file_dict = {
            'sub': self.subjects,
            'ses': self.sessions,
            'task': self.tasks,
            'run': self.runs}
        # Only keep none empty elemenets
        file_types = []
        file_components = []
        for k in ['sub', 'ses', 'task', 'run']:
            if file_dict[k]:
                file_types.append(k)
                file_components += [file_dict[k]]
        file_list = list(itertools.product(*file_components))
        # Specify main directory
        mdir = self.BIDS_dir + '/derivatives/' + self.pipeline
        found_files = []

        for f in file_list:
            wdir = str(mdir)
            fstr = ''
            for i,k in enumerate(file_types):
                if k == 'sub' or k == 'ses':
                    wdir += '/' + k + '-' + f[i] + '/'
                if k != 'sub':
                    fstr += '_'
                else:
                    wdir += 'func/'
                fstr += k + '-' + f[i] + '.*'
            wdir += '/' + self.pipeline_subdir + '/'
            if not self.space:
                space = ''
            else:
                space = '_space-' + self.space

            r = re.compile('^' + fstr + '.*' + space + '.*' + self.last_analysis_step + '.*')
            if os.path.exists(wdir):
                # make filenames
                found = list(filter(r.match, os.listdir(wdir)))
                # Include only if all analysis step tags are present
                found = [i for i in found if all(x in i for x in self.analysis_steps)]
                # Exclude if confounds tag is present
                found = [i for i in found if '_confounds' not in i]
                # Make full paths
                found = list(map(str.__add__,[re.sub('/+','/',wdir)]*len(found),found))

                if found:
                    found_files += found

            if quiet==-1:
                print(wdir)

        if quiet == 0:
            print(found_files)
        return found_files


    def get_confound_files(self,quiet=0):
        """
        Parameters
        ----------
        quiet: int
            If 1, prints results. If 0, no results printed.

        Returns
        -------
        found_files : list
            The files which are currently selected wihch will be used if removing confounds.

            This is specified by confound_pipeline (or pipeline if unset), confound_pipeline_subdir (or pipeline_subdir if unset), space, parcellation, tasks, runs, subjects etc.
        """
        # This could be mnade better
        file_dict = {
            'sub': self.subjects,
            'ses': self.sessions,
            'task': self.tasks,
            'run': self.runs}
        # Only keep none empty elemenets
        file_types = []
        file_components = []
        for k in ['sub', 'ses', 'task', 'run']:
            if file_dict[k]:
                file_types.append(k)
                file_components += [file_dict[k]]
        file_list = list(itertools.product(*file_components))
        # Specify main directory
        if self.confound_pipeline:
            mdir = self.BIDS_dir + '/derivatives/' + self.confound_pipeline
        else:
            mdir = self.BIDS_dir + '/derivatives/' + self.pipeline
        found_files = []

        for f in file_list:
            wdir = str(mdir)
            fstr = ''
            for i,k in enumerate(file_types):
                if k == 'sub' or k == 'ses':
                    wdir += '/' + k + '-' + f[i] + '/'
                if k != 'sub':
                    fstr += '_'
                else:
                    wdir += 'func/'
                fstr += k + '-' + f[i] + '.*'
            wdir_pipesub = wdir + '/' + self.pipeline_subdir + '/'
            # Allow for pipeline_subdir to not be there (ToDo: perhaps add confound_pipeline_subdir in future)
            if os.path.exists(wdir_pipesub):
                wdir = wdir_pipesub
            r = re.compile('^' + fstr + '.*' + '_confounds' + '.*')
            if os.path.exists(wdir):
                found = list(filter(r.match, os.listdir(wdir)))
                found = list(map(str.__add__,[re.sub('/+','/',wdir)]*len(found),found))
                if found:
                    found_files += found


        if quiet == 0:
            print(found_files)
        return found_files



    def get_confound_alternatives(self,quiet=0):
        # This could be mnade better
        file_list = self.get_confound_files(quiet=1)

        confounds = []
        for f in file_list:
            file_format = f.split('.')[-1]
            if  file_format == 'tsv':
                confounds += list(pd.read_csv(f,delimiter='\t').keys())
            elif file_format == 'csv':
                confounds += list(pd.read_csv(f,delimiter=',').keys())

        confounds = sorted(list(set(confounds)))

        if quiet == 0:
            print('Confounds in confound files: \n - ' + '\n - '.join(confounds))
        return confounds

    def set_bad_subjects(self,bad_subjects):

        if isinstance(bad_subjects,str):
            bad_subjects = [bad_subjects]

        for bs in bad_subjects:
            if bs in self.subjects:
                self.subjects.remove(bs)
            else:
                print('WARNING: subject: ' + str(bs) + ' is not found in TenetoBIDS.subjects')

        if not self.bad_subjects:
            self.bad_subjects = bad_subjects
        else:
            self.bad_subjects += bad_subjects


    def set_confound_pipeline(self,confound_pipeline):
        """
        There may be times when the pipeline is updated (e.g. teneto) but you want the confounds from the preprocessing pipieline (e.g. fmriprep).
        To do this, you set the confound_pipeline to be the preprocessing pipeline where the confound files are.

        Parameters
        ----------

        confound_pipeline : str
            Directory in the BIDS_dir where the confounds file is.


        """

        self.add_history(inspect.stack()[0][3], locals(), 1)

        if not os.path.exists(self.BIDS_dir + '/derivatives/' + confound_pipeline):
            print('Specified direvative directory not found.')
            self.get_pipeline_alternatives()
        else:
            # Todo: perform check that pipeline is valid
            self.confound_pipeline = confound_pipeline



    def set_confounds(self,confounds,quiet=0):
        # This could be mnade better

        self.add_history(inspect.stack()[0][3], locals(), 1)

        file_list = self.get_confound_files(quiet=1)

        if isinstance(confounds,str):
            confounds = [confounds]

        for f in file_list:
            file_format = f.split('.')[-1]
            if  file_format == 'tsv':
                sub_confounds = list(pd.read_csv(f,delimiter='\t').keys())
            elif file_format == 'csv':
                sub_confounds = list(pd.read_csv(f,delimiter=',').keys())
            for c in confounds:
                if c not in sub_confounds:
                    print('Warning: the confound (' + c + ') not found in file: ' + f)

        self.confounds = confounds

    def set_network_communities(self,parcellation):
        """

        parcellation : str
            path to csv or name of default parcellation.

        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        # Sett if seperate subcortical atlas is specified
        if '+' in parcellation:
            # Need to add subcortical info to network_communities and network_communities_info_
            parcin = parcellation.split('+')
            parcellation = parcin[0]
            subcortical = parcin[1]
        else:
            subcortical = None

        net_path = teneto.__path__[0] + '/data/parcellation_defaults/' + parcellation + '_network.csv'
        if os.path.exists(parcellation):
            self.network_communities_ = pd.read_csv(parcellation,index_col=0)
            self.network_communities_info_ = self.network_communities_.drop_duplicates().sort_values('network_id').reset_index(drop=True)
            self.network_communities_info_['number_of_nodes'] = self.network_communities_.groupby('network_id').count()
        elif os.path.exists(net_path):
            self.network_communities_ = pd.read_csv(net_path,index_col=0)
            self.network_communities_info_ = self.network_communities_.drop_duplicates().sort_values('network_id').reset_index(drop=True)
            self.network_communities_info_['number_of_nodes'] = self.network_communities_.groupby('network_id').count()
        else:
            print('No (static) network community file found.')

        if subcortical:
            #Assuming only OH atlas exists for subcortical at the moment.
            node_num=21
            sub = pd.DataFrame(data={'Community': ['Subcortical (OH)']*node_num,'network_id':np.repeat(self.network_communities_['network_id'].max()+1,node_num)})
            self.network_communities_ = self.network_communities_.append(sub)
            self.network_communities_.reset_index(drop=True,inplace=True)

    def make_parcellation(self,parcellation,parc_type=None,parc_params=None,network='defaults',update_pipeline=True,removeconfounds=False,tag=None):

        """
        Reduces the data from voxel to parcellation space. Files get saved in a teneto folder in the derivatives with a roi tag at the end.

        Parameters
        -----------

        parcellation : str
            specify which parcellation that you would like to use. For MNI: 'power2012_264', 'gordon2014_333'. TAL: 'shen2013_278'
        parc_type : str
            can be 'sphere' or 'region'. If nothing is specified, the default for that parcellation will be used.
        parc_params : dict
            **kwargs for nilearn functions
        network : str
            if "defaults", it selects static parcellation, _if available_ (other options will be made available soon).
        removeconfounds : bool
            if true, regresses out confounds that are specfied in self.set_confounds with linear regression.
        update_pipeline : bool
            TenetoBIDS gets updated with the parcellated files being selected.
        tag : str
            If multiple types of analysis are going to be run, you can set tag to add a '_tag' on the file name.

        Returns
        -------
        Files are saved in ./BIDS_dir/derivatives/teneto_<version>/.../parcellation/.
        To load these files call TenetoBIDS.load_parcellation.

        NOTE
        ----
        These functions make use of nilearn. Please cite nilearn if used in a publicaiton.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)

        parc_name = parcellation.split('_')[0].lower()

        # Check confounds have been specified
        if not self.confounds and removeconfounds:
            raise ValueError('Specified confounds are not found. Make sure that you have run self.set_confunds([\'Confound1\',\'Confound2\']) first.')

        # In theory these should be the same. So at the moment, it goes through each element and checks they are matched.
        # A matching algorithem may be needed if cases arise where this isnt the case
        files = self.get_selected_files(quiet=1)
        if removeconfounds:
            confound_files = self.get_confound_files(quiet=1)
            if len(files) != len(confound_files):
                print('WARNING: number of confound files does not equal number of selected files')
            for n in range(len(files)):
                if confound_files[n].split('_confounds')[0] not in files[n]:
                    raise ValueError('Confound matching with data did not work.')

        self.set_network_communities(parcellation)

        if not tag:
            tag = ''
        else:
            tag = '_' + tag

        for i,f in enumerate(files):

            file_name = f.split('/')[-1].split('.')[0]
            save_name = file_name + '_parc-' + parc_name + tag + '_roi'
            paths_post_pipeline = f.split(self.pipeline)
            if self.pipeline_subdir:
                paths_post_pipeline = paths_post_pipeline[1].split(self.pipeline_subdir)
            paths_post_pipeline = paths_post_pipeline[1].split(file_name)[0]
            save_dir = self.BIDS_dir + '/derivatives/' + 'teneto_' + teneto.__version__ + '/' + paths_post_pipeline + '/parcellation/'
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            roi = teneto.utils.make_parcellation(f,parcellation,parc_type,parc_params)
            # Make nodd, time
            roi = roi.transpose()

            # Confounds need to be loaded here.
            if removeconfounds:
                if confound_files[i].split('.')[-1] == 'csv':
                    delimiter = ','
                elif confound_files[i].split('.')[-1] == 'tsv':
                    delimiter = '\t'
                df = pd.read_csv(confound_files[i],sep=delimiter)
                df = df[self.confounds]
                if df.isnull().any().any():
                    # Not sure what is the best way to deal with this.
                    # The time points could be ignored. But if multiple confounds, this means these values will get ignored
                    print('WARNING: Some confounds were NaNs. Setting these values to median of confound.')
                    df = df.fillna(df.median())
                patsy_str_confounds = ' + '.join(self.confounds)
                # Linear regresion to regress out (i.e. perform regression and keep residuals) or confound variables.
                for r in range(roi.shape[0]):
                    # Create dataframe
                    df['y'] = roi[r,:]
                    # Specify model
                    model = smf.ols(formula = 'y ~ ' + patsy_str_confounds,data=df)
                    # Fit model
                    res = model.fit()
                    # Get residuals
                    roi[r,:] = res.resid_pearson


            np.save(save_dir + save_name + '.npy', roi)

        if update_pipeline == True:
            if not self.confound_pipeline and len(self.get_confound_files(quiet=1)) > 0:
                self.set_confound_pipeline(self.pipeline)
            self.set_pipeline('teneto_' + teneto.__version__)
            self.set_pipeline_subdir('parcellation')
            self.analysis_steps += self.last_analysis_step
            if tag:
                self.analysis_steps += [tag[1:]]
            self.set_last_analysis_step('roi')
            self.parcellation = parcellation



    def set_last_analysis_step(self,last_analysis_step):
        """
        The last analysis step is the final tag that is present in files.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        self.last_analysis_step = last_analysis_step

    def set_analysis_steps(self,analysis_step,add_step=False):
        """
        Specify which analysis steps are part of the selected files.

        Parameters
        -----------

        analysis_step : str or list
            Analysis tags that are found in the file names of interest. E.g. 'preproc' will only select files with 'preproc' in them.
        add_step : Bool
            If true, then anything in self.analysis_steps is already kept.

        Returns
        -------
        TenetoBIDS.analysis_steps gets updated.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        if isinstance(analysis_step,str):
            if add_step:
                self.analysis_steps.append()
            else:
                self.analysis_steps = [analysis_step]
        elif isinstance(analysis_step,list):
            if add_step:
                self.analysis_steps += analysis_step
            else:
                self.analysis_steps = analysis_step

        else:
            raise ValueError('Invalud input')



    def set_pipeline(self,pipeline):
        """
        Specify the pipeline. See get_pipeline_alternatives to see what are avaialble. Input should be a string.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        if not os.path.exists(self.BIDS_dir + '/derivatives/' + pipeline):
            print('Specified direvative directory not found.')
            self.get_pipeline_alternatives()
        else:
            # Todo: perform check that pipeline is valid
            self.pipeline = pipeline

    def set_pipeline_subdir(self,pipeline_subdir):
        self.add_history(inspect.stack()[0][3], locals(), 1)
#        if not os.path.exists(self.BIDS_dir + '/derivatives/' + self.pipeline + '/' + pipeline_subdir):
#            print('Specified direvative sub-directory not found.')
#            self.get_pipeline_subdir_alternatives()
#        else:
#            # Todo: perform check that pipeline is valid
        self.pipeline_subdir = pipeline_subdir


    def set_runs(self,runs):
        """
        Specify the runs which all selected files must include. See get_run_alternatives to see what are avaialble. Input can be string or list of strings.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        if isinstance(runs,str):
            runs=[runs]
        if self.raw_data_exists:
            runs_in_dataset = self.BIDS.get_runs()
            if len(set(runs).intersection(runs_in_dataset))==len(runs):
                self.runs = sorted(list(runs))
            else:
                raise ValueError('Specified run(s) not founds in BIDS dataset')
        else:
            self.runs = sorted(list(tasks))

    def set_sessions(self,sessions):
        """
        Specify the sessions which all selected files must include. See get_session_alternatives to see what are avaialble.  Input can be string or list of strings.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        if isinstance(sessions,str):
            sessions=[sessions]
        if self.raw_data_exists:
            sessions_in_dataset = self.BIDS.get_sessions()
            if len(set(sessions).intersection(sessions_in_dataset))==len(sessions):
                self.sessions = sorted(list(sessions))
            else:
                raise ValueError('Specified session(s) not founds in BIDS dataset')
        else:
            self.sessions = sorted(list(tasks))

    def set_space(self,space):
        """
        Specify the space which all selected files must include. See get_space_alternatives to see what are avaialble.  Input can be string or list of strings.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)

        space_alternatives = self.get_space_alternatives(quiet=1)
        if space not in space_alternatives:
            raise ValueError('Specified space cannot be found for any subjects. Run TN.get_space_alternatives() to see the optinos in directories.')
        self.space = space

    def set_subjects(self,subjects=None):
        """
        Specify the subjects which are selected files for the analysis.   Input can be string or list of strings.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        if isinstance(subjects,str):
            subjects=[subjects]
        # GEt from raw data or from derivative structure
        if self.raw_data_exists:
            subjects_in_dataset = self.BIDS.get_subjects()
            if len(set(subjects).intersection(subjects_in_dataset))==len(subjects):
                self.subjects = sorted(list(subjects))
            else:
                raise ValueError('Specified subject(s) not founds in BIDS dataset')
        elif not self.raw_data_exists:
            if not self.pipeline:
                raise ValueError('Pipeline must be set if raw_data_exists = False')
            elif not subjects:
                subjects_in_dataset = os.listdir(self.BIDS_dir + '/derivatives/' + self.pipeline)
                subjects_in_dataset = [f.split('sub-')[1] for f in subjects_in_dataset if os.path.isdir(self.BIDS_dir + '/derivatives/' + self.pipeline + '/' + f)]
                self.subjects = subjects_in_dataset
            else:
                self.subjects = subjects



    def set_tasks(self,tasks):
        """
        Specify the space which all selected files must include. See get_task_alternatives to see what are avaialble.  Input can be string or list of strings.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        if isinstance(tasks,str):
            tasks=[tasks]
        if self.raw_data_exists:
            tasks_in_dataset = self.BIDS.get_tasks()
            if len(set(tasks).intersection(tasks_in_dataset))==len(tasks):
                self.tasks = sorted(list(tasks))
            else:
                raise ValueError('Specified task(s) not founds in BIDS dataset')
        else:
            self.tasks = sorted(list(tasks))

    def print_dataset_summary(self):

        """
        Prints information about the the BIDS data and the files currently selected.
        """

        print('--- DATASET INFORMATION ---')

        print('--- Subjects ---')
        if self.raw_data_exists:
            if self.BIDS.get_subjects():
                print('Number of subjects (in dataset): ' + str(len(self.BIDS.get_subjects())))
                print('Subjects (in dataset): ' + ', '.join(self.BIDS.get_subjects()))
            else:
                print('NO SUBJECTS FOUND (is the BIDS directory specified correctly?)')

        print('Number of subjects (selected): ' + str(len(self.subjects)))
        print('Subjects (selected): ' + ', '.join(self.subjects))
        if isinstance(self.bad_subjects,list):
            print('Bad subjects: ' + ', '.join(self.bad_subjects))
        else:
            print('Bad subjects: 0')

        print('--- Tasks ---')
        if self.raw_data_exists:
            if self.BIDS.get_tasks():
                print('Number of tasks (in dataset): ' + str(len(self.BIDS.get_tasks())))
                print('Tasks (in dataset): ' + ', '.join(self.BIDS.get_tasks()))
        if self.tasks:
            print('Number of tasks (selected): ' + str(len(self.tasks)))
            print('Tasks (selected): ' + ', '.join(self.tasks))
        else:
            print('No task names found')

        print('--- Runs ---')
        if self.raw_data_exists:
            if self.BIDS.get_runs():
                print('Number of runs (in dataset): ' + str(len(self.BIDS.get_runs())))
                print('Runs (in dataset): ' + ', '.join(self.BIDS.get_runs()))
        if self.runs:
            print('Number of runs (selected): ' + str(len(self.runs)))
            print('Rubs (selected): ' + ', '.join(self.runs))
        else:
            print('No run names found')


        print('--- Sessions ---')
        if self.raw_data_exists:
            if self.BIDS.get_sessions():
                print('Number of runs (in dataset): ' + str(len(self.BIDS.get_sessions())))
                print('Sessions (in dataset): ' + ', '.join(self.BIDS.get_sessions()))
        if self.sessions:
            print('Number of sessions (selected): ' + str(len(self.sessions)))
            print('Sessions (selected): ' + ', '.join(self.sessions))
        else:
            print('No session names found')

        print('--- PREPROCESSED DATA (Pipelines/Derivatives) ---')

        if not self.pipeline:
            print('Derivative pipeline not set. To set, run TN.set_pipeline()')
        else:
            print('Pipeline: ' + self.pipeline)
        if self.pipeline_subdir:
            print('Pipeline subdirectories: ' + self.pipeline_subdir)
        if not self.space:
            print('Space not set. To set, run TN.set_space()')
        else:
            print('Space: ' + self.space)
        if not self.parcellation:
            print('No parcellation specified. To set, run TN.set_parcellation()')
        else:
            print('Parcellation: ' + self.parcellation)

        selected_files = self.get_selected_files(quiet=1)
        if selected_files:
            print('--- SELECTED DATA ---')
            print('Numnber of selected files: ' + str(len(selected_files)))
            print('\n - '.join(selected_files))

    def save_aspickle(self, fname):
        if fname[-4:] != '.pkl':
            fname += '.pkl'
        with open(fname, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load_frompickle(cls,fname):
        if fname[-4:] != '.pkl':
            fname += '.pkl'
        with open(fname, 'rb') as f:
            return pickle.load(f)


    def load_parcellation_data(self,parcellation=None,tag=None):
        """
        Function returns the data created by. The default grabs all data in the teneto/../func/parcellation directory.

        **INPUT**

        parcellation : str
            Specify parcellation (optional). Default will grab everything that can be found.

        tag : str
            Any additional tag to filter out which files to load.

        **RETURNS**

        :parcellation_data_: numpy array containing the parcellation data. Each file is appended to the first dimension of the numpy array.
        :parcellation_trialinfo_: pandas data frame containing the subject info (all BIDS tags) in the numpy array.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        data_list=[]
        trialinfo_list = []
        if parcellation:
            parc = parcellation
        if not self.parcellation:
            parc = ''
        else:
            parc = self.parcellation.split('_')[0]

        if not tag:
            tag = ''

        for s in self.subjects:
            # Define base folder
            base_path = self.BIDS_dir + '/derivatives/' + self.pipeline
            base_path += '/sub-' + s + '/func/parcellation/'
            file_list=os.listdir(base_path)
            for f in file_list:
                # Include only if all analysis step tags are present
                if parc in f and tag in f:
                    # Get all BIDS tags. i.e. in 'sub-AAA', get 'sub' as key and 'AAA' as item.
                    bid_tags=re.findall('[a-zA-Z]*-',f)
                    bids_tag_dict = {}
                    for t in bid_tags:
                        key = t[:-1]
                        bids_tag_dict[key]=re.findall(t+'[A-Za-z0-9.,*+]*',f)[0].split('-')[-1]
                    if f.split('.')[-1] == 'npy':
                        data = np.load(base_path+f)
                        data_list.append(data)
                        trialinfo = pd.DataFrame(bids_tag_dict,index=[0])
                        trialinfo_list.append(trialinfo)
                    else:
                        print('Warning: Could not find data for a subject')

            self.parcellation_data_ = np.array(data_list)
            if trialinfo_list:
                out_trialinfo = pd.concat(trialinfo_list)
                out_trialinfo.reset_index(inplace=True,drop=True)
                self.parcellation_trialinfo_ = out_trialinfo



    def load_tvc_data(self,tag=None):
        """
        Function loads time-varying connectivity estimates created by the TenetoBIDS.derive function.
        The default grabs all data (in numpy arrays) in the teneto/../func/tvc/ directory.
        Data is placed in teneto.tvc_data_

        **INPUT**

        tag : str
            Any additional tag to filter out which files to load.

        **RETURNS**

        tvc_data_ : numpy array
            Containing the parcellation data. Each file is appended to the first dimension of the numpy array.
        tvc_trialinfo_ : pandas data frame
            Containing the subject info (all BIDS tags) in the numpy array.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        data_list=[]
        trialinfo_list = []

        if not tag:
            tag = ''

        for s in self.subjects:
            # Define base folder
            base_path = self.BIDS_dir + '/derivatives/' + self.pipeline
            base_path += '/sub-' + s + '/func/tvc/'
            file_list=os.listdir(base_path)
            for f in file_list:
                # Include only if all analysis step tags are present
                if os.path.isfile(base_path + f) and tag in f:
                    # Get all BIDS tags. i.e. in 'sub-AAA', get 'sub' as key and 'AAA' as item.
                    bid_tags=re.findall('[a-zA-Z]*-',f)
                    bids_tag_dict = {}
                    for t in bid_tags:
                        key = t[:-1]
                        bids_tag_dict[key]=re.findall(t+'[A-Za-z0-9.,*+]*',f)[0].split('-')[-1]
                    if f.split('.')[-1] == 'npy':
                        data = np.load(base_path+f)
                        data_list.append(data)
                        trialinfo = pd.DataFrame(bids_tag_dict,index=[0])
                        trialinfo_list.append(trialinfo)
                    else:
                        print('Warning: Could not find data for a subject (expecting numpy array)')

            self.tvc_data_ = np.array(data_list)
            if trialinfo_list:
                out_trialinfo = pd.concat(trialinfo_list)
                out_trialinfo.reset_index(inplace=True,drop=True)
                self.tvc_trialinfo_ = out_trialinfo



    def load_network_measure(self,measure,timelocked=False,calc=None,tag=None):
        self.add_history(inspect.stack()[0][3], locals(), 1)
        data_list=[]
        trialinfo_list = []

        if not calc:
            calc = ''
        else:
            calc = 'calc-' + calc

        if not tag:
            tag = ''

        for s in self.subjects:
            # Define base folder
            base_path = self.BIDS_dir + '/derivatives/' + self.pipeline
            base_path += '/sub-' + s + '/func/tvc/temporal-network-measures/' + measure + '/'
            measure_sub = measure
            # Add evet_locked folder if thats what is asked for
            if timelocked:
                base_path += 'timelocked/'
                measure_sub = 'timelocked-' + measure_sub
            # Get files
            file_list=os.listdir(base_path)
            # Get tags in filename
            for f in file_list:
                if os.path.isfile(base_path + f):
                    if calc in f and tag in f:
                        bids_tags=re.findall('[a-zA-Z]*-',f)
                        bids_tag_dict = {}
                        for t in bids_tags:
                            key = t[:-1]
                            bids_tag_dict[key]=re.findall(t+'[A-Za-z0-9.,*+]*',f)[0].split('-')[-1]
                        # Get data
                        if f.split('.')[-1] == 'pkl':
                            df = pd.read_pickle(base_path+f)
                            data = df[measure_sub].values
                            trialinfo = df.drop(measure_sub, 1)
                            for k in bids_tag_dict.keys():
                                trialinfo[k] = bids_tag_dict[k]
                            trialinfo_list.append(trialinfo)
                            for d in data:
                                data_list.append(d)
                        elif f.split('.')[-1] == 'npy':
                            data = np.load(base_path+f)
                            data_list.append(data)
                            trialinfo = pd.DataFrame(bids_tag_dict,index=[0])
                            trialinfo_list.append(trialinfo)

                        else:
                            print('Warning: Could not find pickle data')

            self.networkmeasure_ = np.array(data_list)
            if trialinfo_list:
                out_trialinfo = pd.concat(trialinfo_list)
                out_trialinfo.reset_index(inplace=True,drop=True)
                self.trialinfo_ = out_trialinfo



    def make_timelocked_events(self, measure, event_names, event_onsets, toi, calc=None, tag=None):
        """
        Creates time locked time series of <measure>. Measure must have time in its -1 axis.

        Parameters
        -----------

        measure : str
            temporal network measure that should already exist in the teneto/[subject]/tvc/network-measures directory
        event_names : str or list
            what the event is called (can be list of multiple event names)
        event_onsets: list
            List of onset times (can be list of list for multiple events)
        toi : array
            +/- time points around each event. So if toi = [-10,10] it will take 10 time points before and 10 time points after
        calc : str
            type of network measure calculation.
        tag : str
            any additional tag placed on file name.

        Note
        ----

        Currently no ability to loop over more than one measure

        Returns
        -------
        Creates a time-locked output placed in BIDS_dir/derivatives/teneto_<version>/../tvc/temporal-network-measures/<networkmeasure>/timelocked/
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        event_onsets_combined = list(itertools.chain.from_iterable(event_onsets))
        event_names_list = [[e]*len(event_onsets[i]) for i,e in enumerate(event_names)]
        event_names_list = list(itertools.chain.from_iterable(event_names_list))
        #time_index = np.arange(toi[0],toi[1]+1)

        if not calc:
            calc = ''
        else:
            calc = 'calc-' + calc

        if not tag:
            tag = ''

        for s in self.subjects:

            base_path = self.BIDS_dir + '/derivatives/' + self.pipeline
            base_path += '/sub-' + s + '/func/tvc/temporal-network-measures/' + measure + '/'

            if not os.path.exists(base_path):
                print('Warning: cannot find data for subject: ' + s)

            if not os.path.exists(base_path + '/timelocked/'):
                os.makedirs(base_path + '/timelocked/')

            for f in os.listdir(base_path):
                if os.path.isfile(base_path + f):
                    if calc in f:
                        self_measure = np.load(base_path + '/' + f)
                        # make time dimensions the first dimension
                        self_measure = self_measure.transpose([len(self_measure.shape)-1] + list(np.arange(0,len(self_measure.shape)-1)))
                        tl_data = []
                        for e in event_onsets_combined:
                            tmp = self_measure[e+toi[0]:e+toi[1]+1]
                            # Make time dimension last dimension
                            tmp = tmp.transpose(list(np.arange(1,len(self_measure.shape))) + [0])
                            tl_data.append(tmp)
                        df=pd.DataFrame(data={'timelocked-' + measure: tl_data, 'event': event_names_list, 'event_osnet': event_onsets_combined})
                        net_event = []
                        for e in df['event'].unique():
                            edf = df[df['event']==e]
                            net_tmp = []
                            for r in edf.iterrows():
                                net_tmp.append(r[1]['timelocked-' + measure])
                            net_event.append(np.array(net_tmp).mean(axis=0))
                        # Save output
                        save_dir_base = base_path + 'timelocked/'
                        file_name = f.split('/')[-1].split('.')[0] + '_timelocked'
                        df.to_pickle(save_dir_base + file_name + '.pkl')


    def load_participant_data(self):
        """
        Loads the participanets.tsv file that is placed in BIDS_dir as participants_.
        """
        self.add_history(inspect.stack()[0][3], locals(), 1)
        self.participants_ = pd.read_csv(self.BIDS_dir + 'participants.tsv',delimiter='\t')



def temporal_community(G,omega=1,gamma=1,randomseed=None):

    #Inititialization
    np.random.seed(randomseed)
    #for t in np.arange()
    B = np.zeros(np.shape(G))

    for g in range(0,G.shape[-1]):
        C = louvain(G[:,:,g],randomseed)
        Bs = np.zeros(G[:,:,g].shape)
        for c in np.unique(C):
            fid = np.where(C==c)
            Bs[np.ix_(fid[0],fid[0])]=1
        B[:,:,g] = Bs

    #B[np.arange(0,3),np.arange(3,6)]=1

    # This create a string of the block matrix that is to be created.
    strbuild = '['
    for g in range(0,B.shape[-1]):
        strbuild += '['
        strbuild += 'None,'*(g-1)
        if g != 0:
            strbuild += 'np.array(np.matlib.identity(B.shape[0])*omega),'
        strbuild += 'B[:,:,' + str(g) + '],'
        if g != B.shape[-1]-1:
            strbuild +='np.array(np.matlib.identity(B.shape[0])*omega),'
        strbuild += 'None,'*(B.shape[-1]-2-g)
        strbuild += '],'
    strbuild += ']'
    if B.shape[-1]>2:
        Bbmat = sp.sparse.bmat(eval(strbuild))
        Gsp=sp.sparse.csr_matrix(Bbmat)
    elif B.shape[-1]==2:
        Gden = eval(strbuild)
        Gden = np.vstack([np.hstack(Gden[0]),np.hstack(Gden[1])])
        Gsp=sp.sparse.csr_matrix(Gden)
    else:
        raise ValueError('Only one layer detected')

    C=louvain(Gsp)
    C=C.reshape(G.shape[-1],G.shape[0]).transpose()
    return C



"""
networkmeasures: intercontacttimes
"""

import numpy as np
import teneto.utils as utils


def intercontacttimes(netin):
    """
    Calculates the intercontacttimes of each edge in a network

    **PARAMETERS**

    :netin: Temporal network (craphlet or contact).

        :nettype: 'bu', 'bd'

    **OUTPUT**

    :contacts: intercontact times as numpy array

        :format: dictionary

    **NOTES**

    Connections are assumed to be binary

    **SEE ALSO**

    *bursty_coeff*

    **History**

    :Modified: Dec 2016, WHT
    :Created: Nov 2016, WHT

    """

    # Process input
    netin, netinfo = utils.process_input(netin, ['C', 'G', 'TO'])

    if netinfo['nettype'][0] == 'd':
        print('WARNING: assuming connections to be binary when computing intercontacttimes')

    # Each time series is padded with a 0 at the start and end. Then t[0:-1]-[t:].
    # Then discard the noninformative ones (done automatically)
    # Finally return back as np array
    contacts = np.array([[None] * netinfo['netshape'][0]] * netinfo['netshape'][1])

    if netinfo['nettype'][1] == 'u':
        for i in range(0, netinfo['netshape'][0]):
            for j in range(i + 1, netinfo['netshape'][0]):
                edge_on = np.where(netin[i, j, :] > 0)[0]
                edge_on = np.append(0, edge_on)
                edge_on = np.append(edge_on, 0)
                edge_on_diff = edge_on[2:-1] - edge_on[1:-2]
                contacts[i, j] = np.array(edge_on_diff)
                contacts[j, i] = np.array(edge_on_diff)
    elif netinfo['nettype'][1] == 'd':
        for i in range(0, netinfo['netshape'][0]):
            for j in range(0, netinfo['netshape'][0]):
                edge_on = np.where(netin[i, j, :] > 0)[0]
                edge_on = np.append(0, edge_on)
                edge_on = np.append(edge_on, 0)
                edge_on_diff = edge_on[2:-1] - edge_on[1:-2]
                contacts[i, j] = np.array(edge_on_diff)

    out = {}
    out['intercontacttimes'] = contacts
    out['nettype'] = netinfo['nettype']
    return out



# Main function to draw a slice_graph
import matplotlib.pyplot as plt
import numpy as np
import math
from teneto.utils import *
from teneto.plot.slice_plot import make_bezier, pascal_row


def circle_plot(netIn, ax, nlabs=[], linestyle='k-', nodesize=1000):
    '''

    Function draws "circle plot" and exports axis handles


    **PARAMETERS**

    :netIn: temporal network input (graphlet or contact)
    :ax: matplotlib ax handles.
    :nlabs: nodes labels. List of strings
    :linestyle: bezier line style
    :nodesize: size of nodes


    **OUTPUT**

    :ax: axis handle of slice graph


    **SEE ALSO**

    - *slice_plot*
    - *graphlet_stack_plot*


    **HISTORY**

    :updated: Dec 2016, WHT
    :created: Sept 2016, WHT

    '''
    # Get input type (C or G)
    inputType = checkInput(netIn, conMat=1)
    nettype = 'xx'
    # Convert C representation to G
    if inputType == 'M':
        shape = np.shape(netIn)
        edg = np.where(np.abs(netIn) > 0)
        contacts = [tuple([edg[0][i], edg[1][i]])
                    for i in range(0, len(edg[0]))]
        netIn = {}
        netIn['contacts'] = contacts
        netIn['netshape'] = shape
    elif inputType == 'G':
        netIn = graphlet2contact(netIn)
        inputType = 'C'

    if inputType == 'C':
        edgeList = [tuple(np.array(e[0:2]) + e[2] * netIn['netshape'][0])
                    for e in netIn['contacts']]
    elif inputType == 'M':
        edgeList = netIn['contacts']

    n = netIn['netshape'][0]
    # Get positions of node on unit circle
    posx = [math.cos((2 * math.pi * i) / n) for i in range(0, n)]
    posy = [math.sin((2 * math.pi * i) / n) for i in range(0, n)]
    # Get Bezier lines in a circle
    for edge in edgeList:
        bvx, bvy = bezier_circle(
            (posx[edge[0]], posy[edge[0]]), (posx[edge[1]], posy[edge[1]]), 20)
        ax.plot(bvx, bvy, linestyle)
    ax.scatter(posx, posy, s=nodesize, c=range(0, n))
    # Remove things that make plot unpretty
    ax.set_yticklabels([])
    ax.set_xticklabels([])
    ax.set_yticks([])
    ax.set_xticks([])
    ax.set_frame_on(False)
    # make plot a square
    x0, x1 = ax.get_xlim()
    y0, y1 = ax.get_ylim()
    ax.set_aspect((x1 - x0) / (y1 - y0))
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    return ax


# Adapaption of bezier_points but for going in towards the centre of circle
def bezier_circle(p1, p2, pointN):
    ts = [t / pointN for t in range(pointN + 1)]
#    d=p1[0]-(max(p1[1],p2[1])-min(p1[1],p2[1]))/negxLim
    bezier = make_bezier([p1, (0, 0), p2])
    points = bezier(ts)
    bvx = [i[0] for i in points]
    bvy = [i[1] for i in points]
    return bvx, bvy



"""

Networkmeasures: Temporal Efficiency

"""

import numpy as np
from teneto.networkmeasures.shortest_temporal_path import shortest_temporal_path




def temporal_efficiency(data, calc='global'):
    """
    returns temporal efficiency estimate.


    **PARAMETERS**

    :data: This is either:

        :netIn: temporal network input (graphlet or contact).

            :nettype: 'bu'

        :paths: Dictionary of paths (output of shortest_temporal_path).


    :calc: 'global' (default) - measure averages over time and nodes.
        'node' or 'node_from' average over nodes (i) and time. Giving average efficiency for i to j.
        'node_to' measure average over nodes j and time.
         Giving average efficiency using paths to j from  i.


    **OUTPUT**

    :E: global temporal efficiency (global measure)

        :format: integer (numpy array)

    **NOTES**

    This can be implemented on a non-global level in the future.

    **SEE ALSO**

    - *shortesttemporalpath*

    **HISTORY**

    Modified - Jan 2016, WHT (documentation)
    Created - Dec 2016, WHT

    """

    paths = 0  # are shortest paths calculated
    if isinstance(data, dict):
        # This could be calcne better
        if [k for k in list(data.keys()) if k == 'paths'] == ['paths']:
            paths = 1
    # if shortest paths are not calculated, calculate them
    if paths == 0:
        data = shortest_temporal_path(data)

    # Calculate efficiency which is 1 over the mean path.
    if calc == 'global':
        eff = 1 / np.nanmean(data['paths'])
    elif calc == 'node' or calc == 'node_from':
        eff = 1 / np.nanmean(np.nanmean(data['paths'], axis=2), axis=1)
    elif calc == 'node_to':
        eff = 1 / np.nanmean(np.nanmean(data['paths'], axis=2), axis=0)

    return eff



"""
generatenetwork.rand_binomial
"""

import numpy as np
from teneto.utils import graphlet2contact


def rand_binomial(size, prob, netrep='graphlet', nettype='bu', initialize='zero', netinfo=None):
    """

    Creates a random binary network following a binomial distribution.

    Parameters
    ----------

    size : list or array of length 2 or 3.  
    
        Input [n,t] generates n number of nodes and t number of time points.  
        Can also be of length 3 (node x node x time) but number of nodes in 3-tuple must be identical.

    prob : int or list/array of length 2.  
     
        If int, this indicates probabability for each node becoming active (equal for all nodes).
        
        If tuple/list of length 2, this indicates different probabilities for edges to become active/inactive.

            The first value is "birth rate". The probability of an absent connection becoming present.

            The second value is the "death rate". This dictates the probability of an edge present remaining present.

            example : [40,60] means there is a 40% chance that a 0 will become a 1 and a 60% chance that a 1 stays a 1.

    netrep : str 
        network representation: 'graphlet' (default) or 'contact'.
    nettype : str 
        Weighted or directed network. String 'bu' or 'bd' (accepts 'u' and 'd' as well as b is implicit)
    initialize : float or str 
        Input percentage (in decimal) for how many nodes start activated. Alternative specify 'zero' (default) for all nodes to start deactivated.      
    netinfo : dict 
        Dictionary for contact representaiton information. 
 
    Returns 
    -------

    net : array or dict 
        
        Generated nework. Format depends on netrep input argument.

    Note
    ------

    Option 2 of the "prob" parameter can be used to create a small autocorrelaiton or make sure that, once an edge has been present, it never disapears.

    
    Read more
    ---------

    There is some work on the properties on the graphs with birth/death rates (called edge-Markovian Dynamic graphs) as described here. Clementi et al (2008) Flooding Time in edge-Markovian Dynamic Graphs *PODC*

    """

    size = np.atleast_1d(size)
    prob = np.atleast_1d(prob)
    if len(size) == 2 or (len(size) == 3 and size[0] == size[1]):
        pass
    else:
        raise ValueError('size input should be [numberOfNodes,Time]')
    if len(prob) > 2:
        raise ValueError('input: prob must be of len 1 or len 2')
    if prob.min() < 0 or prob.max() > 1:
        raise ValueError('input: prob should be probability between 0 and 1')
    if nettype[-1] == 'u' or nettype[-1] == 'd':
        pass
    else:
        raise ValueError('nettype must be u or d')

    network_size = size[0]
    nr_time_points = size[-1]
    connmat = network_size * network_size
    if len(prob) == 1:
        net = np.random.binomial(1, prob, connmat * nr_time_points)
        net = net.reshape(network_size * network_size, nr_time_points)
    if len(prob) == 2:
        net = np.zeros([connmat, nr_time_points])
        if initialize == 'zero':
            pass
        else:
            edgesat0 = np.random.randint(
                0, connmat, int(np.round(initialize * (connmat))))
            net[edgesat0, 0] = 1
        for t_ind in range(0, nr_time_points - 1):
            edges_off = np.where(net[:, t_ind] == 0)[0]
            edges_on = np.where(net[:, t_ind] == 1)[0]
            update_edges_on = np.random.binomial(1, prob[0], len(edges_off))
            update_edge_off = np.random.binomial(1, prob[1], len(edges_on))
            net[edges_off, t_ind + 1] = update_edges_on
            net[edges_on, t_ind + 1] = update_edge_off
    # Set diagonal to 0
    net[np.arange(0, network_size * network_size, network_size + 1), :] = 0
    # Reshape to graphlet
    net = net.reshape([network_size, network_size, nr_time_points])
    # only keep upper left if nettype = u
    # Note this could be made more efficient by only doing (network_size*network_size/2-network_size) nodes
    # in connmat and inserted directly into upper triangular.
    if nettype[-1] == 'u':
        unet = np.zeros(net.shape)
        ind = np.triu_indices(network_size)
        unet[ind[0], ind[1], :] = np.array(net[ind[0], ind[1], :])
        unet = unet + np.transpose(unet, [1, 0, 2])
        net = unet
    if netrep == 'contact':
        if not netinfo:
            netinfo = {}
        netinfo['nettype'] = 'b' + nettype[-1]
        net = graphlet2contact(net, netinfo)
    return net



"""

derive: different methods to derive dynamic functional connectivity

"""


import numpy as np
from statsmodels.stats.weightstats import DescrStatsW
import teneto
import scipy.stats as sps


def derive(data, params):
    """
    Derives connectivity from the data. A lot of data is inherently built with edges
     (e.g. communication between two individuals).
    However other networks are derived from the covariance of time series
     (e.g. brain networks between two regions).

    Covariance based metrics deriving time-resolved networks can be done in multiple ways.
     There are other methods apart from covariance based.

    Derive a weight vector for each time point and then the corrrelation coefficient
     for each time point.

    Paramters
    --------

    data : array 
        Time series data to perform connectivity derivation on. (Default dimensions are: (time as rows, nodes as columns). Change params{'dimord'} if you want it the other way (see below).
    
    params : dict 
        Parameters for each method (see below).

    Necessary paramters
    ===================

    method : str
        method: "distance","slidingwindow", "taperedslidingwindow",
     "jackknife", "multiplytemporalderivative". Alternatively, method can be a weight matrix of size time x time.

    **Different methods have method specific paramaters (see below)**

    Params for all methods (optional)
    =================================

    postpro : "no" (default). Other alternatives are: "fisher", "boxcox", "standardize"
     and any combination seperated by a + (e,g, "fisher+boxcox").
      See postpro_pipeline for more information.
    dimord : str
        Dimension order: 'node,time' (default) or 'time,node'. People like to represent their data differently and this is an easy way to be sure that you are inputing the data in the correct way.
    analysis_id : str or int 
        add to identify specfic analysis. Generated report will be placed in './report/' + analysis_id + '/derivation_report.html
    report : str 
        Options: "yes" or "no" (default). A report is saved in ./report/[analysis_id]/derivation_report.html if "yes"
    report_path : str 
        String where the report is saved. Default is ./report/[analysis_id]/derivation_report.html 

    Methods specific parameters 
    =========================== 

    method == "distance"
    ~~~~~~~~~~~~~~~~~~~

    Distance metric calculates 1/Distance metric weights, and scales between 0 and 1.
    W[t,t] is excluded from the scaling and then set to 1.

    params['distance']: str 
        Distance metric (e.g. 'euclidean'). See teneto.utils.getDistanceFunction for more info

    When method == "slidingwindow"
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    params['windowsize'] : int 
        Size of window.

    When method == "taperedslidingwindow"
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    params['windowsize'] : int 
        Size of window.
    params['distribution'] : str 
        Scipy distribution (e.g. 'norm','expon'). Any distribution here: https://docs.scipy.org/doc/scipy/reference/stats.html
    params['distribution_params'] : list 
        Each parameter, excluding the data "x" (in their scipy function order) to generate pdf.

        NOTE
        !!!!!!!!!!
        The data x should be considered to be centered at 0 and have a length of window size.
         (i.e. a window size of 5 entails x is [-2, -1, 0, 1, 2] a window size of 6 entails [-2.5, -1.5, 0.5, 0.5, 1.5, 2.5])
        Given x params['distribution_params'] contains the remaining parameters.

        e.g. normal distribution requires pdf(x, loc, scale) where loc=mean and scale=std.
         This means that the mean and std have to be provided in distribution_params.

        Say we have a gaussian distribution, a window size of 21 and params['distribution_params'] is [0,5].
         This will lead to a gaussian with its peak at in the middle of each window with a standard deviation of 5.

        Instead, if we set params['distribution_params'] is [10,5] this will lead to a half gaussian with its peak at the final time point with a standard deviation of 5.

    When method == "temporalderivative"
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    params['windowsize'] : int
        Size of window.

    When method == "jackknife"
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    No parameters are necessary. 
    
    Optional parameters: 
    
    params['weight-var'] : array, (optional) 
        NxN array to weight the JC estimates (standerdized-JC*W). If weightby is selected, do not standerdize in postpro.
    params['weight-mean'] : array, (optional) 
        NxN array to weight the JC estimates (standerdized-JC+W). If weightby is selected, do not standerdize in postpro.


    Returns
    -------    

    G : array 
        Connectivity estimates (nodes x nodes x time)


    READ MORE
    ---------
    About the general weighted pearson approach used for most methods, see: 
    Thompson & Fransson (2019) A common framework for the problem of deriving estimates of dynamic functional brain connectivity. 
    Neuroimage. (https://doi.org/10.1016/j.neuroimage.2017.12.057)

    SEE ALSO
    --------

    *postpro_pipeline*, *gen_report*

    """
    report = {}

    if 'dimord' not in params.keys():
        params['dimord'] = 'node,time'

    if 'report' not in params.keys():
        params['report'] = 'no'

    if 'analysis_id' not in params.keys():
        params['analysis_id'] = ''

    if 'postpro' not in params.keys():
        params['postpro'] = 'no'

    if params['report'] != 'no':

        if 'analysis_id' not in params.keys():
            params['analysis_id'] = ''

        if 'report_path' not in params.keys():
            params['report_path'] = './report/' + params['analysis_id']

        if 'report_filename' not in params.keys():
            params['report_filename'] = 'derivation_report.html'

    if params['dimord'] == 'node,time':
        data = data.transpose()

    if isinstance(params['method'], str):
        if params['method'] == 'jackknife':
            weights, report = weightfun_jackknife(data.shape[0], report)
            relation = 'weight'
        elif params['method'] == 'sliding window' or params['method'] == 'slidingwindow':
            weights, report = weightfun_sliding_window(
                data.shape[0], params, report)
            relation = 'weight'
        elif params['method'] == 'tapered sliding window' or params['method'] == 'taperedslidingwindow':
            weights, report = weightfun_tapered_sliding_window(
                data.shape[0], params, report)
            relation = 'weight'
        elif params['method'] == 'distance' or params['method'] == "spatial distance" or params['method'] == "node distance" or params['method'] == "nodedistance" or params['method'] == "spatialdistance":
            weights, report = weightfun_spatial_distance(data, params, report)
            relation = 'weight'
        elif params['method'] == 'mtd' or params['method'] == 'multiply temporal derivative' or params['method'] == 'multiplytemporalderivative' or params['method'] == 'temporal derivative' or params['method'] == "temporalderivative":
            R, report = temporal_derivative(data, params, report)
            relation = 'coupling'
        else:
            raise ValueError(
                'Unrecognoized method. See derive_with_weighted_pearson documentation for predefined methods or enter own weight matrix')
    else:
        try:
            weights = np.array(params['method'])
            relation = 'weight'
        except:
            raise ValueError(
                'Unrecognoized method. See documentation for predefined methods')
        if weights.shape[0] != weights.shape[1]:
            raise ValueError("weight matrix should be square")
        if weights.shape[0] != data.shape[0]:
            raise ValueError("weight matrix must equal number of time points")

    if relation == 'weight':
        # Loop over each weight vector and calculate pearson correlation.
        # Note, should see if this can be made quicker in future.
        R = np.array(
            [DescrStatsW(data, weights[i, :]).corrcoef for i in range(0, weights.shape[0])])
        # Make node,node,time
        R = R.transpose([1, 2, 0])

    # Correct jackknife direction
    if params['method'] == 'jackknife':
        # Correct inversion 
        R = R * -1
        jc_z = 0
        if 'weight-var' in params.keys():
            R = np.transpose(R, [2, 0, 1])
            R = (R - R.mean(axis=0)) / R.std(axis=0)
            jc_z = 1
            R = R * params['weight-var']
            R = R.transpose([1,2,0])
        if 'weight-mean' in params.keys():
            R = np.transpose(R, [2, 0, 1])
            if jc_z == 0:
                R = (R - R.mean(axis=0)) / R.std(axis=0)
            R = R + params['weight-mean']
            R = np.transpose(R, [1, 2, 0])
        R = teneto.utils.set_diagonal(R,1)



    if params['postpro'] != 'no':
        R, report = teneto.derive.postpro_pipeline(
            R, params['postpro'], report)
        R = teneto.utils.set_diagonal(R,1)

    if params['report'] == 'yes':
        teneto.derive.gen_report(report, params['report_path'], params['report_filename'])
    return R


def weightfun_jackknife(T, report):
    """ 
    Creates the weights for the jackknife method. See func: teneto.derive.derive. 
    """

    weights = np.ones([T, T])
    np.fill_diagonal(weights, 0)
    report['method'] = 'jackknife'
    report['jackknife'] = ''
    return weights, report


def weightfun_sliding_window(T, params, report):
    """ 
    Creates the weights for the sliding window method. See func: teneto.derive.derive. 
    """
    weightat0 = np.zeros(T)
    weightat0[0:params['windowsize']] = np.ones(params['windowsize'])
    weights = np.array([np.roll(weightat0, i)
                        for i in range(0, T + 1 - params['windowsize'])])
    report['method'] = 'slidingwindow'
    report['slidingwindow'] = params
    report['slidingwindow']['taper'] = 'untapered/uniform'
    return weights, report


def weightfun_tapered_sliding_window(T, params, report):
    """ 
    Creates the weights for the tapered method. See func: teneto.derive.derive. 
    """
    x = np.arange(-(params['windowsize'] - 1) / 2, (params['windowsize']) / 2)
    distribution_parameters = ','.join(map(str, params['distribution_params']))
    taper = eval('sps.' + params['distribution'] +
                 '.pdf(x,' + distribution_parameters + ')')

    weightat0 = np.zeros(T)
    weightat0[0:params['windowsize']] = taper
    weights = np.array([np.roll(weightat0, i)
                        for i in range(0, T + 1 - params['windowsize'])])
    report['method'] = 'slidingwindow'
    report['slidingwindow'] = params
    report['slidingwindow']['taper'] = taper
    report['slidingwindow']['taper_window'] = x
    return weights, report


def weightfun_spatial_distance(data, params, report):
    """ 
    Creates the weights for the spatial distance method. See func: teneto.derive.derive. 
    """
    distance = teneto.utils.getDistanceFunction(params['distance'])
    weights = np.array([distance(data[n, :], data[t, :]) for n in np.arange(
        0, data.shape[0]) for t in np.arange(0, data.shape[0])])
    weights = np.reshape(weights, [data.shape[0], data.shape[0]])
    np.fill_diagonal(weights, np.nan)
    weights = 1 / weights
    weights = (weights - np.nanmin(weights)) / \
        (np.nanmax(weights) - np.nanmin(weights))
    np.fill_diagonal(weights, 1)
    return weights, report


def temporal_derivative(data, params, report):
    """ 
    Performs mtd method. See func: teneto.derive.derive. 
    """
    # Data should be timexnode
    report = {}

    # Derivative
    tdat = data[1:, :] - data[:-1, :]
    # Normalize
    tdat = tdat / np.std(tdat, axis=0)
    # Coupling
    coupling = np.array([tdat[:, i] * tdat[:, j] for i in np.arange(0,
                                                                    tdat.shape[1]) for j in np.arange(0, tdat.shape[1])])
    coupling = np.reshape(
        coupling, [tdat.shape[1], tdat.shape[1], tdat.shape[0]])
    # Average over window using strides
    shape = coupling.shape[:-1] + (coupling.shape[-1] -
                                   params['windowsize'] + 1, params['windowsize'])
    strides = coupling.strides + (coupling.strides[-1],)
    coupling_windowed = np.mean(np.lib.stride_tricks.as_strided(
        coupling, shape=shape, strides=strides), -1)

    report = {}
    report['method'] = 'temporalderivative'
    report['temporalderivative'] = {}
    report['temporalderivative']['windowsize'] = params['windowsize']

    return coupling_windowed, report



"""
networkmeasures.fluctuatbility
"""
import numpy as np
import teneto.utils as utils


def fluctuability(netin, calc='global'):
    """
    fluctuability of temporal networks.This is the variation in unique edges
     through time divided by the overall number of edges.

    **PARAMETERS**

    :netin: temporal network input (graphlet or contact)

        :nettype: 'bd', 'bu', 'wu', 'wd'

    :calc: version of fluctuabiility to calcualte. 'global'
     (i.e. average distance of all nodes for each consecutive time point).
      A nodal version may be added in future.

    **OUTPUT**

    :F: Fluctuability

        :format: scalar (calc='global')

    **SEE ALSO**
    - *voalitility*

    **HISTORY**

    :Modified: Jan 2016, WHT (documentation)
    :Created: Dec 2016, WHT

    """

    # Get input type (C or G)
    netin, netinfo = utils.process_input(netin, ['C', 'G', 'TO'])

    netin[netin != 0] = 1
    unique_edges = np.sum(netin, axis=2)
    unique_edges[unique_edges > 0] = 1
    unique_edges[unique_edges == 0] = 0

    fluct = (np.sum(unique_edges)) / np.sum(netin)
    return fluct




import igraph as ig
import louvain 
from scipy.spatial.distance import jaccard
import itertools
import numpy as np
import teneto 

def temporal_louvain_with_consensus(G, iter_n=100, resolution_parameter=1, interslice_weight=0, quality_function='NewmanGirvan2004', seed=42, consensus_threshold=0.5): 
    """
    Temporal louvain clustering run for iter_n times and consenus matrix returned (no temporal consensus matrix yet)

    Parameters 
    ----------
    G : array 
        graphlet 
    iter_n : int 
        nummber of repeated louvain clustering 
    resolution_parameter : int 
        Spatial resolution parameter. Only valid for some qualtiy functions. Default=1.
        Resolution parameter is only needed for ReichardtBornholdt2006, and   
    interslice_weight : int 
        The weight that connects the different graphlets/snapshots to eachother. Default=0
    quality function : str 
        What type of louvain clustering is done. Options: NewmanGirvan2004, TraagVanDoorenNesterov2011, ReichardtBornholdt2006 
    seed : int 
        Seed for reproduceability 
    consensus_threshold : float 
        Value between 0 and 1. When creating consensus matrix, ignore if value only occurs in specified fraction of iterations. If 0.5 two nodes must be in the same community 50% of the time to be considered in the consensus matrix. 

    Returns 
    -------
    communities : array 
        Consensus matrix from louvain clustering. Dimensions: commiunities, time
    
    Qualtify funciton sources
    --------------------------

    NewmanGirvan2004 : 
        Newman, M. E. J., & Girvan, M. (2004). Finding and evaluating community structure in networks. Physical Review E, 69(2), 026113. 10.1103/PhysRevE.69.026113
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#modularityvertexpartition>`_ 
    ReichardtBornholdt2006 : 
        Reichardt, J., & Bornholdt, S. (2006). Statistical mechanics of community detection. Physical Review E, 74(1), 016110. 10.1103/PhysRevE.74.016110
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#rbconfigurationvertexpartition>`_ 
    TraagVanDoorenNesterov2011 : 
        Traag, V. A., Van Dooren, P., & Nesterov, Y. (2011). Narrow scope for resolution-limit-free community detection. Physical Review E, 84(1), 016114. 10.1103/PhysRevE.84.016114
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#cpmvertexpartition>`_ 
    TraagKringsVanDooren2013 : 
        Traag, V. A., Krings, G., & Van Dooren, P. (2013). Significant scales in community structure. Scientific Reports, 3, 2930. 10.1038/srep02930
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#significancevertexpartition>`_ 
    TraagAldecoaDelvenne2015 : 
        Traag, V. A., Aldecoa, R., & Delvenne, J.-C. (2015). Detecting communities using asymptotical surprise. Physical Review E, 92(2), 022816. 10.1103/PhysRevE.92.022816
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#surprisevertexpartition>`_ 
    
    Dependencies
    ------------ 
    These functions make use of iGraph (http://igraph.org/python/) and louvain-igraph (http://louvain-igraph.readthedocs.io/en/latest/)

    Note 
    ----
    At the moment input should generally only be positive edges. 

    """
    Gin = G.copy()
    D = np.random.random(np.prod(Gin.shape)).reshape(Gin.shape)
    i=0
    # While condition doesn't seem to be run at the moment
    while len(np.where((D > 0) & (D < 1))[0]) > 0:
        i=i+1
        C = teneto.communitydetection.temporal_louvain(Gin, iter_n=iter_n, resolution_parameter=resolution_parameter, seed=seed, interslice_weight=interslice_weight, quality_function=quality_function)
        D = [teneto.communitydetection.make_consensus_matrix(C[:,:,t],th=consensus_threshold) for t in range(C.shape[-1])]
        D = np.dstack(D) 
        Gin = D
        #Only first iteration needs to be returned as consensus means they are all identical 
    communities = teneto.communitydetection.make_temporal_consensus(C[0,:,:])
    return communities 

def temporal_louvain(G, iter_n=1, resolution_parameter=1, interslice_weight=1, quality_function='NewmanGirvan2004', seed=100):
    """
    Temporal louvain clustering run for iter_n times and consenus matrix returned (no temporal consensus matrix yet)

    Parameters 
    ----------
    G : array 
        graphlet 
    iter_n : int 
        nummber of repeated louvain clustering 
    resolution_parameter : int 
        Spatial resolution parameter. Only valid for some qualtiy functions. Default=1.
        Resolution parameter is only needed for ReichardtBornholdt2006, and   
    interslice_weight : int 
        The weight that connects the different graphlets/snapshots to eachother. Default=0
    quality function : str 
        What type of louvain clustering is done. Options: NewmanGirvan2004, TraagVanDoorenNesterov2011, ReichardtBornholdt2006 
    seed : int 
        Seed for reproduceability 
    consensus_threshold : float 
        Value between 0 and 1. When creating consensus matrix, ignore if value only occurs in specified fraction of iterations. If 0.5 two nodes must be in the same community 50% of the time to be considered in the consensus matrix. 

    Returns 
    -------
    communities : array 
        Louvain clustering. Dimensions: [iter_n], commiunities, time
    
    Qualtify funciton sources
    --------------------------

    NewmanGirvan2004 : 
        Newman, M. E. J., & Girvan, M. (2004). Finding and evaluating community structure in networks. Physical Review E, 69(2), 026113. 10.1103/PhysRevE.69.026113
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#modularityvertexpartition>`_ 
    ReichardtBornholdt2006 : 
        Reichardt, J., & Bornholdt, S. (2006). Statistical mechanics of community detection. Physical Review E, 74(1), 016110. 10.1103/PhysRevE.74.016110
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#rbconfigurationvertexpartition>`_ 
    TraagVanDoorenNesterov2011 : 
        Traag, V. A., Van Dooren, P., & Nesterov, Y. (2011). Narrow scope for resolution-limit-free community detection. Physical Review E, 84(1), 016114. 10.1103/PhysRevE.84.016114
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#cpmvertexpartition>`_ 
    TraagKringsVanDooren2013 : 
        Traag, V. A., Krings, G., & Van Dooren, P. (2013). Significant scales in community structure. Scientific Reports, 3, 2930. 10.1038/srep02930
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#significancevertexpartition>`_ 
    TraagAldecoaDelvenne2015 : 
        Traag, V. A., Aldecoa, R., & Delvenne, J.-C. (2015). Detecting communities using asymptotical surprise. Physical Review E, 92(2), 022816. 10.1103/PhysRevE.92.022816
        `Read more <http://louvain-igraph.readthedocs.io/en/latest/reference.html#surprisevertexpartition>`_ 
    
    Dependencies
    ------------ 
    These functions make use of iGraph (http://igraph.org/python/) and louvain-igraph (http://louvain-igraph.readthedocs.io/en/latest/)

    Note 
    ----
    At the moment input should generally only be positive edges. 



    """
    if quality_function == 'TraagVanDoorenNesterov2011':
        louvain_alg = louvain.CPMVertexPartition
        louvain_kwags = {'resolution_parameter': resolution_parameter}
    elif quality_function == 'ReichardtBornholdt2006': 
        louvain_alg = louvain.RBConfigurationVertexPartition 
        louvain_kwags = {'resolution_parameter': resolution_parameter}
    elif quality_function == 'NewmanGirvan2004': 
        louvain_alg = louvain.ModularityVertexPartition
        louvain_kwags = {}
    elif quality_function == 'TraagKringsVanDooren2013': 
        louvain_alg = louvain.SignificanceVertexPartition
        louvain_kwags = {}
    elif quality_function == 'TraagAldecoaDelvenne2015': 
        louvain_alg = louvain.SurpriseVertexPartition
        louvain_kwags = {}
    g_to_ig = []
    for i in range(G.shape[-1]): 
        g_to_ig.append(ig.Graph.Weighted_Adjacency(G[:,:,i].tolist()))
    for n in range(G.shape[0]):
        for t in range(G.shape[-1]):
            g_to_ig[t].vs[n]['id'] = n
    membership = []
    louvain.set_rng_seed(seed)
    if interslice_weight != 0: 
        for n in range(0,iter_n):
            mem, improvement = louvain.find_partition_temporal(
                    g_to_ig,
                    louvain_alg,
                    interslice_weight=interslice_weight,
                    **louvain_kwags)
            membership.append(mem)
        com_membership = np.array(membership).transpose([0,2,1])
    else:
        com_membership = [] 
        for n in range(0,iter_n):
            membership = [] 
            for snapshot in g_to_ig:
                mem = louvain.find_partition(
                    snapshot,
                    louvain_alg,
                    **louvain_kwags)   
                membership.append(mem.membership)
            com_membership.append(membership)
        com_membership = np.array(com_membership).transpose([0,2,1])
    return np.squeeze(com_membership)

def make_consensus_matrix(com_membership,th=0.5): 
    """
    Makes the consensus matrix
.
    Parameters
    ----------

    com_membership : array 
        Shape should be iterations,node. 

    th : float 
        threshold to cancel noisey edges  

    Returns 
    ------- 

    D : array 
        consensus matrix 
    """
    com_membership = np.array(com_membership)
    D = np.zeros([com_membership.shape[1],com_membership.shape[1]])
    for it in range(com_membership.shape[0]):
        for c in np.unique(com_membership[it,:]):
            id = np.where(com_membership[it,:]==c)
            if len(id[0]) > 1: 
                id = np.array(list(itertools.combinations(id[0],2)))
                D[id[:,0],id[:,1]] += 1
                D[id[:,1],id[:,0]] += 1
    D = D/com_membership.shape[0]
    D[D<th] = 0
    np.fill_diagonal(D,1)
    return D

def make_temporal_consensus(communities): 
    """ 
    Links communities where communities have been derived at multiple time points. 

    Parameters 
    --------- 
    """
    for t in range(1,communities.shape[-1]): 
        community_possibilities = []
        jaccard_best = 1
        cidx = list(itertools.permutations(np.unique(communities[:,t]),len(np.unique(communities[:,t]))))
        for c in cidx: 
            ctmp = communities[:,t].copy()
            for i,n in enumerate(c):
                ctmp[communities[:,t]==i]=n
            if jaccard(communities[:,t-1],ctmp) < jaccard_best:
                community_possibilities = ctmp
        communities[:,t] = community_possibilities
    return communities




"""
File contains functions for postprocessing derivation of connectivity estimates
"""

import numpy as np
import scipy as sp
import teneto


def postpro_fisher(data, report=None):
    """
    Performs fisher transform on everything in data.

    If report variable is passed, this is added to the report.
    """
    if not report:
        report = {}
    # Due to rounding errors
    data[data < -0.99999999999999] = -1
    data[data > 0.99999999999999] = 1
    fisher_data = 0.5 * np.log((1 + data) / (1 - data))
    report['fisher'] = {}
    report['fisher']['performed'] = 'yes'
    #report['fisher']['diagonal'] = 'zeroed'
    return fisher_data, report


def postpro_boxcox(data, report=None):
    """
    Performs box cox transform on everything in data.

    If report variable is passed, this is added to the report.
    """
    if not report:
        report = {}
    # Note the min value of all time series will now be at least 1.
    mindata = 1 - np.nanmin(data)
    data = data + mindata
    ind = np.triu_indices(data.shape[0], k=1)

    boxcox_list = np.array([sp.stats.boxcox(np.squeeze(
        data[ind[0][n], ind[1][n], :])) for n in range(0, len(ind[0]))])

    boxcox_data = np.zeros(data.shape)
    boxcox_data[ind[0], ind[1], :] = np.vstack(boxcox_list[:, 0])
    boxcox_data[ind[1], ind[0], :] = np.vstack(boxcox_list[:, 0])

    bccheck = np.array(np.transpose(boxcox_data, [2, 0, 1]))
    bccheck = (bccheck - bccheck.mean(axis=0)) / bccheck.std(axis=0)
    bccheck = np.squeeze(np.mean(bccheck, axis=0))
    np.fill_diagonal(bccheck, 0)

    report['boxcox'] = {}
    report['boxcox']['performed'] = 'yes'
    report['boxcox']['lambda'] = [
        tuple([ind[0][n], ind[1][n], boxcox_list[n, -1]]) for n in range(0, len(ind[0]))]
    report['boxcox']['shift'] = mindata
    report['boxcox']['shited_to'] = 1

    if np.sum(np.isnan(bccheck)) > 0:
        report['boxcox'] = {}
        report['boxcox']['performed'] = 'FAILED'
        report['boxcox']['failure_reason'] = (
            'Box cox transform is returning edges with uniform values through time. '
            'This is probabaly due to one or more outliers or a very skewed distribution. '
            'Have you corrected for sources of noise (e.g. movement)? '
            'If yes, some time-series might need additional transforms to approximate to Gaussian.'
        )
        report['boxcox']['failure_consequence'] = (
            'Box cox transform was skipped from the postprocess pipeline.'
        )
        boxcox_data = data - mindata
        error_msg = ('TENETO WARNING: Box Cox transform problem. \n'
                     'Box Cox transform not performed. \n'
                     'See report for more details.')
        print(error_msg)

    return boxcox_data, report


def postpro_standardize(data, report=None):
    """
    Standardizes everything in data (along axis -1).

    If report variable is passed, this is added to the report.
    """
    if not report:
        report = {}
    # First make dim 1 = time.
    data = np.transpose(data, [2, 0, 1])
    standardized_data = (data - data.mean(axis=0)) / data.std(axis=0)
    standardized_data = np.transpose(standardized_data, [1, 2, 0])
    report['standardize'] = {}
    report['standardize']['performed'] = 'yes'
    report['standardize']['method'] = 'Z-score'
    # The above makes self connections to nan, set to 1. 
    data = teneto.utils.set_diagonal(data,1)
    return standardized_data, report


def postpro_pipeline(data, pipeline, report=None):
    """

    :PARAMETERS:

    :data: pearson correlation values in temporal matrix form (node,node,time)
    :pipeline: list or string (if string, each steps seperated by + sign).

        :options: 'fisher','boxcox','standardize'

        Each of the above 3 can be specified. If fisher is used, it must be before boxcox.
        If standardize is used it must be after boxcox and fisher.

    :report: default is empty. If non-empty, appended to report.

    :OUTPUT:

    :prepro_data: postprocessed data
    :postprocessing_info: dictionary of information about postprocessing
     (e.g lambda parameters for boxcox)

    """

    postpro_functions = {
        'fisher': teneto.derive.postpro_fisher,
        'boxcox': teneto.derive.postpro_boxcox,
        'standardize': teneto.derive.postpro_standardize
    }

    if not report:
        report = {}

    if isinstance(pipeline, str):
        pipeline = pipeline.split('+')

    report['postprocess'] = []
    for postpro_step in pipeline:
        report['postprocess'].append(postpro_step)
        prepro_data, report = postpro_functions[postpro_step](data, report)
    return prepro_data, report



"""

Networkmeasures: shortest temporal paths

"""

import numpy as np
import teneto.utils as utils


def shortest_temporal_path(netin, quiet=1):
    """
    Calculates the shortest temporal path when all possible routes can be travelled
     at each time point.
    Currently only works for binary undirected edges (but can be expanded).

    **PARAMETERS**

    :netin: temporal network input (graphlet or contact)

        :nettype: 'bu'

    :quiet: quiet (default = 1). Turn to 0 if you want progree update.

    **OUTPUT**

    :paths_dict: shortest temporal paths

        :format: dictionary

    Paths are of the struction, path['paths'][i,j,t] - shortest path for i to reach j,
     starting at time t.

    **NOTE**

    This function assumes all paths can be taken per time point.
    In a future update, this function temporalPaths will allow for
     only a portion of edges to be travelled
     per time point.
    This will be implmeneted with no change to the funcitonality of calling
     this function as it is today,
     with the defaults being all edges can be travelled.

    **SEE ALSO**

    - *temporal_efficiency*
    - *reachability_latency*
    - *temporal_closeness_centrality*

    **HISTORY**

    Modified - Aug 2017, WHT (PEP8)
    Modified - Dec 2016, WHT (documentation)
    Created - Nov 2016, WHT

    """

    # Get input type (C or G)
    # Process input
    netin, netinfo = utils.process_input(netin, ['C', 'G', 'TO'])

    if netinfo['nettype'] != 'bu':
        errormsg = ('It looks like your graph is not binary and undirected. '
                    'Shortest temporal paths can only be calculated for '
                    'binary undirected networks in Teneto at the moment. '
                    'If another type is required, please create an issue at: '
                    'github.com/wiheto/teneto.')
        raise ValueError(errormsg)

    # Preallocate output
    paths = np.zeros(netinfo['netshape']) * np.nan
    # Go backwards in time and see if something is reached
    paths_last_contact = np.zeros(
        [netinfo['netshape'][0], netinfo['netshape'][1]]) * np.nan
    for t_ind in list(reversed(range(0, netinfo['netshape'][2]))):
        if quiet == 0:
            print('--- Running for time: ' + str(t_ind) + ' ---')
        fid = np.where(netin[:, :, t_ind] >= 1)
        # Update time step
        # Note to self: Add a conditional to prevent nan warning
        # that can pop out if no path is there straight away.
        paths_last_contact += 1
        # Reset connections present to 1s
        paths_last_contact[fid[0], fid[1]] = 1
        # Update nodes with no connections
        # Nodes to update are nodes with an edge present at the time point
        for i in np.unique(fid[0]):
            connections = np.where(paths_last_contact[i, :] == 1)[0]
            #paths_last_contact_preupdate = np.array(paths_last_contact[i,:])
            paths_last_contact[i, :] = np.nanmin(
                paths_last_contact[np.hstack([connections, i]), :], axis=0)
            # make self connection nan regardless
            paths_last_contact[i, i] = np.nan
        paths[:, :, t_ind] = paths_last_contact
    # Return output
    paths_dict = {}
    percentreach = (paths.size - np.sum(np.isnan(paths))) / \
        (paths.size - (paths.shape[0] * paths.shape[2]))
    paths_dict['percentReached'] = percentreach
    paths_dict['paths'] = paths
    paths_dict['nettype'] = netinfo['nettype']

    return paths_dict
